# wolfTPM (TPM 2.0)

Portable TPM 2.0 project designed for embedded use.


## Project Features

* This implementation provides all TPM 2.0 API’s in compliance with the specification.
* This uses the TPM Interface Specification (TIS) to communicate over SPI.
* The design allows for easy portability to different platforms:
	* Native C code designed for embedded use.
	* Single IO callback for hardware SPI interface.
	* No external dependencies.
	* Compact code size and minimal memory use.
* Examples for the Raspberry Pi and STM32 with CubeMX.
* Includes example code for most TPM2 native API’s.
* Includes wrappers for Key Generation, RSA encrypt/decrypt, ECC sign/verify and ECDH.
* Testing done using the Infineon OPTIGA SLB9670 module and LetsTrust TPM for Raspberry Pi.


## TPM 2.0 Overview

### Hierarchies

```
Platform    TPM_RH_PLATFORM
Owner       TPM_RH_OWNER
Endorsement TPM_RH_ENDORSEMENT
```

Each hierarchy has their own manufacture generated seed.

The arguments used on `TPM2_Create` or `TPM2_CreatePrimary` create a template, which is fed into a KDF to produce the same key based hierarchy used. The key generated is the same each time; even after reboot. The generation of a new RSA 2048 bit key takes about 15 seconds. Typically these are created and then stored in NV using `TPM2_EvictControl`. Each TPM generates their own keys uniquely based on the seed.

There is also an Ephemeral hierarchy (`TPM_RH_NULL`), which can be used to create ephemeral keys.

### Platform Configuration Registers (PCRs)

Contains hash digests for SHA-1 and SHA-256 with an index 0-23. These hash digests can be extended to prove the integrity of a boot sequence (secure boot).


### Terminology

This project uses the terms append vs. marshall and parse vs. unmarshall.



## Platform

This example was written for use on Raspberry Pi® 3 or the STM32 with the CubeMX HAL. This was tested using the 

The Raspberry 3 uses the native `spi_dev` interface and defaults to `/dev/spidev0.1`. If you are running the Infineon patches it overrides the kernel SPI interface with their `spi_tis_dev`, which currently causes this demo to fail.

This has only been tested and confirmed working with Rasbian 4.4.x.

### SPI IO Callback

For interfacing to your hardware platform see the example `tpm2_demo.c` callback function `TPM2_IoCb`. Here you can modify or insert your own IO callback code for the TPM demo.


### Hardware

Tested with:

* Infineon OPTIGA (TM) Trusted Platform Module 2.0 SLB 9670.
* LetsTrust: http://letstrust.de (https://buyzero.de/collections/andere-platinen/products/letstrust-hardware-tpm-trusted-platform-module). Compact Raspberry Pi TPM 2.0 board based on Infineon SLB 9670.


## Building

Build wolfSSL:

```
./autogen.sh
./configure --enable-ecc --enable-sha512 && make
sudo make install
sudo ldconfig
```

Build wolfTPM:

```
./autogen.sh
./configure && make
./examples/wrap/wrap_test
./examples/native/native_test
```


## Release Notes

### wolfTPM Release 1.1 (03/09/2018)

* Added TPM2 wrapper layer to simplify key creation, RSA encrypt/decrypt, ECC sign/verify and ECDH.
* Added TPM2 wrapper example code.
* Added Linux SPI support for running on Raspberry Pi.
* Fixes for TPM2 command and response assembly and parsing.
* Fixes to support authentication for command and response.
* Progress on supporting parameter encryption/decryption.
* Refactor of TIS and Packet layers into new files.
* Fixes/improvements to `wolfTPM2_GetRCString` for error code and string reporting.
* Added new `TPM2_Cleanup` function.
* New tests for TPM2 native API's (test coverage is about 75%).

### wolfTPM Release 1.0 (02/06/2018)

* Support for all TPM2 native API's using TIS and SPI IO callback.
* Helper for getting TPM return code string `TPM2_GetRCString`.
* TPM 2.0 demo code in `examples/tpm/tpm2_demo.c` with support for STM32 CubeMX SPI as reference.


## Examples


### TPM2 Wrapper Tests

```
./examples/wrap/wrap_test
TPM2 Demo for Wrapper API's
RSA Encrypt Test Passed
ECC Sign/Verify Passed
ECC DH Generation Passed
NV Test on index 0x1800200 with 1024 bytes passed
```

### TPM2 Benchmarks

Run on Infineon OPTIGA SLB9670:

```
./examples/bench/bench
TPM2 Benchmark using Wrapper API's
RSA     2048 Public        65 ops took 1.005 sec, avg 15.466 ms, 64.657 ops/sec
RSA     2048 Private        3 ops took 1.343 sec, avg 447.759 ms, 2.233 ops/sec
RSA     2048 Pub  OAEP     12 ops took 1.040 sec, avg 86.657 ms, 11.540 ops/sec
RSA     2048 Priv OAEP      2 ops took 1.032 sec, avg 515.885 ms, 1.938 ops/sec
ECDSA    256 sign          14 ops took 1.037 sec, avg 74.101 ms, 13.495 ops/sec
ECDSA    256 verify         8 ops took 1.027 sec, avg 128.417 ms, 7.787 ops/sec
ECDHE    256 agree          8 ops took 1.040 sec, avg 130.003 ms, 7.692 ops/sec
```

### TPM Native Tests

```
./examples/native/native_test
TPM2 Demo using Native API's
TPM2: Caps 0x30000697, Did 0x001b, Vid 0x15d1, Rid 0x10 
TPM2_Startup pass
TPM2_SelfTest pass
TPM2_GetTestResult: Size 10, Rc 0x0
TPM2_IncrementalSelfTest: Rc 0x0, Alg 0x1 (Todo 0)
TPM2_GetCapability: Property FamilyIndicator 0x322e3000
TPM2_GetCapability: Property PCR Count 24
TPM2_GetRandom: Got 32 bytes
TPM2_StirRandom: success
TPM2_PCR_Read: Index 0, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 1, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 2, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 3, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 4, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 5, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 6, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 7, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 8, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 9, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 10, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 11, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 12, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 13, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 14, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 15, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 16, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 17, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 18, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 19, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 20, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 21, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 22, Digest Sz 32, Update Counter 21
TPM2_PCR_Read: Index 23, Digest Sz 32, Update Counter 21
TPM2_PCR_Extend success
TPM2_PCR_Read: Index 0, Digest Sz 32, Update Counter 22
TPM2_StartAuthSession: sessionHandle 0x3000000
TPM2_PolicyGetDigest: size 32
TPM2_PCR_Read: Index 0, Digest Sz 20, Update Counter 22
wc_Hash of PCR[0]: size 32
TPM2_PolicyPCR failed 0x1c4: TPM_RC_AUTHSIZE
TPM2_PolicyPCR: Updated
TPM2_PolicyRestart: Done
TPM2_HashSequenceStart: sequenceHandle 0x80000000
Hash SHA256 test success
TPM2_CreatePrimary: Endorsement 0x80000000 (314 bytes)
TPM2_CreatePrimary: Storage 0x80000001 (282 bytes)
TPM2_LoadExternal: 0x80000002
TPM2_MakeCredential: credentialBlob 68, secret 256
TPM2_ReadPublic Handle 0x80000002: pub 314, name 34, qualifiedName 34
Create HMAC-SHA256 Key success, public 48, Private 141
TPM2_Load New HMAC Key Handle 0x80000002
TPM2_PolicyCommandCode: success
TPM2_ObjectChangeAuth failed 0x9a2: TPM_RC_BAD_AUTH
TPM2_ObjectChangeAuth: private 2
TPM2_ECC_Parameters: CurveID 3, sz 256, p 32, a 32, b 32, gX 32, gY 32, n 32, h 1
TPM2_Create: New ECDSA Key: pub 88, priv 126
TPM2_Load ECDSA Key Handle 0x80000002
TPM2_Sign: ECC S 32, R 32
TPM2_VerifySignature: Tag 32802
TPM2_Create: New ECDH Key: pub 88, priv 126
TPM2_Load ECDH Key Handle 0x80000002
TPM2_ECDH_KeyGen: zPt 68, pubPt 68
TPM2_Create: New RSA Key: pub 278, priv 222
TPM2_Load RSA Key Handle 0x80000002
TPM2_RSA_Encrypt: 256
TPM2_RSA_Decrypt: 32
RSA Encrypt/Decrypt test passed
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE
TPM2_NV_DefineSpace failed 0x184: TPM_RC_VALUE
```


### TPM2 Wrapper Tests with Debug Enabled

Built with `./configure --enable-debug` or `#define DEBUG_WOLFTPM`

```
./examples/wrap/wrap_test
TPM2 Demo for Wrapper API's
TPM2: Caps 0x30000697, Did 0x001b, Vid 0x15d1, Rid 0x10 
Command: 12
	80 01 00 00 00 0c 00 00 01 44 00 00             | .........D..
Response: 10
	80 01 00 00 00 0a 00 00 01 00                   | ..........
TPM2_Startup pass
Command: 14
	80 01 00 00 00 0e 00 00 01 73 81 00 02 00       | .........s....
Response: 366
	80 01 00 00 01 6e 00 00 00 00 01 1a 00 01 00 0b | .....n..........
	00 03 04 72 00 00 00 06 00 80 00 43 00 10 08 00 | ...r.......C....
	00 00 00 00 01 00 af 17 14 f3 dd 71 f5 4b ce 09 | ...........q.K..
	04 40 18 30 25 18 97 4e 7d 97 e2 6a 99 7f 1c 79 | .@.0%..N}..j...y
	d8 f1 bc eb 97 f1 6a 63 6d 43 60 a1 30 5a fc 14 | ......jcmC`.0Z..
	b5 e3 d1 e0 b7 39 90 43 30 11 8c e7 01 09 e8 01 | .....9.C0.......
	a8 bd e9 60 08 6d 8c a3 c1 a0 a5 40 e6 56 dc 98 | ...`.m.....@.V..
	84 75 4a ca 69 17 d9 0e f7 66 70 ce 64 51 20 b8 | .uJ.i....fp.dQ .
	70 c0 d5 86 ad b4 81 ab bd f2 43 73 3f 8c 2a 39 | p.........Cs?.*9
	f1 3a 21 18 82 c3 1d d3 39 d6 73 84 51 90 d3 f0 | .:!.....9.s.Q...
	7a 08 cc a7 f4 1c 6a 6f 27 48 43 bb ed af 3c a7 | z.....jo'HC...<.
	9c 6e da b7 12 04 28 14 07 23 72 c3 01 e6 c6 c9 | .n....(..#r.....
	b6 ff 86 3d c2 4e dc 7c 66 1f 62 0b 88 32 26 19 | ...=.N.|f.b..2&.
	0c cd 72 63 9c aa 39 ef 87 6b d1 2f 2f 3a 03 96 | ..rc..9..k.//:..
	7e 34 b9 06 1a 4f 6e f2 f3 0f 9f 4c 33 37 35 e0 | ~4...On....L375.
	93 f6 be 31 5c 6f b3 83 50 88 57 71 31 9f 1d 57 | ...1\o..P.Wq1..W
	fb 9b 4b 8d 5d c1 66 b1 be ea f6 5e 00 15 91 13 | ..K.].f....^....
	76 ab c8 6b e4 ad 86 a9 57 ad fb e2 2c 45 c5 90 | v..k....W...,E..
	11 cc 6f bd 5e f3 00 22 00 0b 23 24 f6 f5 3c 45 | ..o.^.."..#$..<E
	70 7c 3e 0a d7 78 3e bc 01 ae cb d9 73 0a 54 49 | p|>..x>.....s.TI
	92 14 63 05 c6 a8 16 ed fc fc 00 22 00 0b 46 d7 | ..c........"..F.
	59 62 84 e1 eb e4 2d 48 d7 0c 35 33 ae 63 81 23 | Yb....-H..53.c.#
	f2 e4 36 e0 da 08 00 f4 6f 96 91 76 99 c8       | ..6.....o..v..
TPM2_ReadPublic Handle 0x81000200: pub 282, name 34, qualifiedName 34
Command: 356
	80 02 00 00 01 64 00 00 01 53 81 00 02 00 00 00 | .....d...S......
	00 1f 40 00 00 09 00 00 01 00 16 54 68 69 73 49 | ..@........ThisI
	73 4d 79 53 74 6f 72 61 67 65 4b 65 79 41 75 74 | sMyStorageKeyAut
	68 00 13 00 0f 54 68 69 73 49 73 4d 79 4b 65 79 | h....ThisIsMyKey
	41 75 74 68 00 00 01 16 00 01 00 0b 00 06 04 60 | Auth...........`
	00 00 00 10 00 10 08 00 00 00 00 00 01 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00                                     | ....
Response: 714
	80 02 00 00 02 ca 00 00 00 00 00 00 02 b7 00 de | ................
	00 20 d3 8f 69 fa 94 75 1e 1b 69 26 c5 78 6e 0e | . ..i..u..i&.xn.
	4d 84 e9 f1 e3 5b 5c 38 41 fb 0e 19 05 8b 73 b4 | M....[\8A.....s.
	0c ff 00 10 6f ad 78 90 af bd 45 ee e2 5f ce 07 | ....o.x...E.._..
	0a 51 dc 7d 42 4e d6 3d 0e e7 8f 23 36 ef c2 ab | .Q.}BN.=...#6...
	98 d9 3b 05 58 d0 b3 15 88 46 37 03 a6 f4 6f fc | ..;.X....F7...o.
	a6 2e 84 af d6 d0 9c ea 71 d0 bc f1 1c ba 29 12 | ........q.....).
	ad 0b fd 16 53 fe 5a 00 06 36 b9 55 35 35 6a 73 | ....S.Z..6.U55js
	2e 01 5e 9b ec a0 76 52 3a 79 8e 1a bb e8 b3 c0 | ..^...vR:y......
	de 15 bb ef 7d a7 bb c1 aa 1d 5e 01 85 69 02 97 | ....}.....^..i..
	b5 56 aa 44 35 d3 80 00 9c 5c 7d 98 eb 36 b5 73 | .V.D5....\}..6.s
	da 1a 18 8c 38 52 15 88 e4 d4 81 7e 4d 28 92 2e | ....8R.....~M(..
	dd 76 ff 89 c9 dc 51 ba e4 af db 68 64 99 44 a9 | .v....Q....hd.D.
	0c 84 1f 0b 86 26 29 10 4e 90 28 eb 17 56 ed 84 | .....&).N.(..V..
	cc bf f4 35 d7 ca 1a 62 67 7b 64 2a e4 90 01 16 | ...5...bg{d*....
	00 01 00 0b 00 06 04 60 00 00 00 10 00 10 08 00 | .......`........
	00 00 00 00 01 00 e6 9c 5f f4 fb 88 26 be 55 2b | ........_...&.U+
	ce 8e a2 32 b4 a7 93 5a 83 30 6a bf 6f a9 09 ab | ...2...Z.0j.o...
	bf 9a 25 e3 83 29 dd b5 87 cf b1 02 85 b5 44 1a | ..%..)........D.
	7f 41 02 c0 72 b2 5e 63 ef 96 a3 0e 03 8f 05 26 | .A..r.^c.......&
	41 6b d2 99 5a a6 89 a8 66 8f c7 3c 76 c8 be 65 | Ak..Z...f..<v..e
	8e 15 37 8b 85 71 a2 93 f8 75 42 68 39 4e bc 4c | ..7..q...uBh9N.L
	46 c9 40 17 bb 92 ec 3d 11 cd 44 40 31 62 13 5f | F.@....=..D@1b._
	1a b9 80 70 b7 58 57 bf 95 c4 76 42 fa 4a b7 cc | ...p.XW...vB.J..
	c0 b0 b2 93 3d 1b d8 ab 92 10 ee a6 8d cd c0 bf | ....=...........
	76 3b c7 1a ac 6a cf 44 14 3a 8a a1 f6 e0 d6 69 | v;...j.D.:.....i
	b1 44 dc c5 43 43 cd 8c db f5 07 db bf 5e 5d 34 | .D..CC.......^]4
	cb dc c0 dd f5 35 f6 6e c4 4b c9 28 d8 86 ac 81 | .....5.n.K.(....
	12 1d bd 09 3e d8 9a af 2b ec 64 b1 65 fa 32 3b | ....>...+.d.e.2;
	84 f8 a4 e2 20 e4 72 a4 03 0d 5e a7 0c 2e 7c d6 | .... .r...^...|.
	03 3a 59 83 9c 10 cd 25 1a 4e 48 21 1f f5 55 0d | .:Y....%.NH!..U.
	11 dd a9 1f 34 c7 f4 94 fb b4 ee 88 e5 4c 6d 28 | ....4........Lm(
	80 b9 6d b3 d8 2f 00 73 00 00 00 00 00 20 e3 b0 | ..m../.s..... ..
	c4 42 98 fc 1c 14 9a fb f4 c8 99 6f b9 24 27 ae | .B.........o.$'.
	41 e4 64 9b 93 4c a4 95 99 1b 78 52 b8 55 01 00 | A.d..L....xR.U..
	0b 00 22 00 0b 23 24 f6 f5 3c 45 70 7c 3e 0a d7 | .."..#$..<Ep|>..
	78 3e bc 01 ae cb d9 73 0a 54 49 92 14 63 05 c6 | x>.....s.TI..c..
	a8 16 ed fc fc 00 22 00 0b 46 d7 59 62 84 e1 eb | ......"..F.Yb...
	e4 2d 48 d7 0c 35 33 ae 63 81 23 f2 e4 36 e0 da | .-H..53.c.#..6..
	08 00 f4 6f 96 91 76 99 c8 00 00 00 20 31 13 9d | ...o..v..... 1..
	8e 56 51 c7 7b 71 e0 0c 82 ea 60 5b ca e3 eb 54 | .VQ.{q....`[...T
	32 ad f6 e1 9b 94 bb fa c9 b0 b6 8c 59 80 21 40 | 2...........Y.!@
	00 00 01 00 20 d9 46 03 f5 7a ce 52 08 b6 a6 8d | .... .F..z.R....
	44 90 2a de 03 2c c2 bd 59 8b 89 37 af 91 40 17 | D.*..,..Y..7..@.
	aa cc 29 ed fe 00 00 01 00 00                   | ..).......
TPM2_Create key: pub 278, priv 222
Command: 553
	80 02 00 00 02 29 00 00 01 57 81 00 02 00 00 00 | .....)...W......
	00 1f 40 00 00 09 00 00 01 00 16 54 68 69 73 49 | ..@........ThisI
	73 4d 79 53 74 6f 72 61 67 65 4b 65 79 41 75 74 | sMyStorageKeyAut
	68 00 de 00 20 d3 8f 69 fa 94 75 1e 1b 69 26 c5 | h... ..i..u..i&.
	78 6e 0e 4d 84 e9 f1 e3 5b 5c 38 41 fb 0e 19 05 | xn.M....[\8A....
	8b 73 b4 0c ff 00 10 6f ad 78 90 af bd 45 ee e2 | .s.....o.x...E..
	5f ce 07 0a 51 dc 7d 42 4e d6 3d 0e e7 8f 23 36 | _...Q.}BN.=...#6
	ef c2 ab 98 d9 3b 05 58 d0 b3 15 88 46 37 03 a6 | .....;.X....F7..
	f4 6f fc a6 2e 84 af d6 d0 9c ea 71 d0 bc f1 1c | .o.........q....
	ba 29 12 ad 0b fd 16 53 fe 5a 00 06 36 b9 55 35 | .).....S.Z..6.U5
	35 6a 73 2e 01 5e 9b ec a0 76 52 3a 79 8e 1a bb | 5js..^...vR:y...
	e8 b3 c0 de 15 bb ef 7d a7 bb c1 aa 1d 5e 01 85 | .......}.....^..
	69 02 97 b5 56 aa 44 35 d3 80 00 9c 5c 7d 98 eb | i...V.D5....\}..
	36 b5 73 da 1a 18 8c 38 52 15 88 e4 d4 81 7e 4d | 6.s....8R.....~M
	28 92 2e dd 76 ff 89 c9 dc 51 ba e4 af db 68 64 | (...v....Q....hd
	99 44 a9 0c 84 1f 0b 86 26 29 10 4e 90 28 eb 17 | .D......&).N.(..
	56 ed 84 cc bf f4 35 d7 ca 1a 62 67 7b 64 2a e4 | V.....5...bg{d*.
	90 01 16 00 01 00 0b 00 06 04 60 00 00 00 10 00 | ..........`.....
	10 08 00 00 00 00 00 01 00 e6 9c 5f f4 fb 88 26 | ..........._...&
	be 55 2b ce 8e a2 32 b4 a7 93 5a 83 30 6a bf 6f | .U+...2...Z.0j.o
	a9 09 ab bf 9a 25 e3 83 29 dd b5 87 cf b1 02 85 | .....%..).......
	b5 44 1a 7f 41 02 c0 72 b2 5e 63 ef 96 a3 0e 03 | .D..A..r.^c.....
	8f 05 26 41 6b d2 99 5a a6 89 a8 66 8f c7 3c 76 | ..&Ak..Z...f..<v
	c8 be 65 8e 15 37 8b 85 71 a2 93 f8 75 42 68 39 | ..e..7..q...uBh9
	4e bc 4c 46 c9 40 17 bb 92 ec 3d 11 cd 44 40 31 | N.LF.@....=..D@1
	62 13 5f 1a b9 80 70 b7 58 57 bf 95 c4 76 42 fa | b._...p.XW...vB.
	4a b7 cc c0 b0 b2 93 3d 1b d8 ab 92 10 ee a6 8d | J......=........
	cd c0 bf 76 3b c7 1a ac 6a cf 44 14 3a 8a a1 f6 | ...v;...j.D.:...
	e0 d6 69 b1 44 dc c5 43 43 cd 8c db f5 07 db bf | ..i.D..CC.......
	5e 5d 34 cb dc c0 dd f5 35 f6 6e c4 4b c9 28 d8 | ^]4.....5.n.K.(.
	86 ac 81 12 1d bd 09 3e d8 9a af 2b ec 64 b1 65 | .......>...+.d.e
	fa 32 3b 84 f8 a4 e2 20 e4 72 a4 03 0d 5e a7 0c | .2;.... .r...^..
	2e 7c d6 03 3a 59 83 9c 10 cd 25 1a 4e 48 21 1f | .|..:Y....%.NH!.
	f5 55 0d 11 dd a9 1f 34 c7 f4 94 fb b4 ee 88 e5 | .U.....4........
	4c 6d 28 80 b9 6d b3 d8 2f                      | Lm(..m../
Response: 59
	80 02 00 00 00 3b 00 00 00 00 80 00 00 01 00 00 | .....;..........
	00 24 00 22 00 0b d3 a4 8f b8 14 46 40 2b f5 b1 | .$.".......F@+..
	19 c2 0d 60 be 15 26 2d b8 9d d4 3a b5 44 1e a4 | ...`..&-...:.D..
	3b 6f 64 a4 2d cd 00 00 01 00 00                | ;od.-......
TPM2_Load Key Handle 0x80000001
Command: 54
	80 01 00 00 00 36 00 00 01 74 80 00 00 01 00 20 | .....6...t..... 
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	00 17 00 0b 00 00                               | ......
Response: 268
	80 01 00 00 01 0c 00 00 00 00 01 00 59 b1 2a be | ............Y.*.
	91 dd e4 23 8c d1 b2 46 48 95 ac fe 1a db 2a a9 | ...#...FH.....*.
	83 bc b7 2a 42 60 38 c6 22 39 e4 be 1d f4 89 3d | ...*B`8."9.....=
	93 dd 7d c4 bd d7 a9 1a c3 98 25 9f 39 23 62 ee | ..}.......%.9#b.
	bc b6 02 90 fa 8a 11 39 23 59 a3 53 7e 3e 60 ff | .......9#Y.S~>`.
	4d 99 53 d7 73 c9 80 a0 c5 f4 54 37 20 d3 e1 84 | M.S.s.....T7 ...
	31 b6 f4 91 c2 7b 66 08 54 02 fd 5b d7 30 77 3d | 1....{f.T..[.0w=
	45 6c bb 37 ec d5 2e 9c 1e fe 20 7d ac 69 03 cb | El.7...... }.i..
	68 d7 d0 6c 27 17 4f 95 03 d0 54 c3 b7 e0 3c 7b | h..l'.O...T...<{
	7d 69 39 bb 77 0f fd 0e fe 87 bb 93 9e e8 4c 13 | }i9.w.........L.
	a2 60 1c a4 ec 99 64 62 fc 68 c6 42 ea 8a 5d 1a | .`....db.h.B..].
	4a 01 4f 0a ef be 7b 74 fb b5 f9 92 1a 90 c7 10 | J.O...{t........
	c8 ce a2 5b 87 43 6a 22 00 d4 9d 58 45 8f 23 84 | ...[.Cj"...XE.#.
	1b 9b f8 5b 44 ef e4 81 c5 75 f1 e2 1f 53 1a d0 | ...[D....u...S..
	c6 60 40 d4 1b 57 e6 fe b8 ff 79 21 37 41 b8 8e | .`@..W....y!7A..
	14 0a 8f 6e c6 67 68 27 e1 8a ed e5 ae 35 09 35 | ...n.gh'.....5.5
	da 56 68 56 0e 36 ee 47 86 26 42 66             | .VhV.6.G.&Bf
TPM2_RSA_Encrypt: 256
Command: 306
	80 02 00 00 01 32 00 00 01 59 80 00 00 01 00 00 | .....2...Y......
	00 18 40 00 00 09 00 00 01 00 0f 54 68 69 73 49 | ..@........ThisI
	73 4d 79 4b 65 79 41 75 74 68 01 00 59 b1 2a be | sMyKeyAuth..Y.*.
	91 dd e4 23 8c d1 b2 46 48 95 ac fe 1a db 2a a9 | ...#...FH.....*.
	83 bc b7 2a 42 60 38 c6 22 39 e4 be 1d f4 89 3d | ...*B`8."9.....=
	93 dd 7d c4 bd d7 a9 1a c3 98 25 9f 39 23 62 ee | ..}.......%.9#b.
	bc b6 02 90 fa 8a 11 39 23 59 a3 53 7e 3e 60 ff | .......9#Y.S~>`.
	4d 99 53 d7 73 c9 80 a0 c5 f4 54 37 20 d3 e1 84 | M.S.s.....T7 ...
	31 b6 f4 91 c2 7b 66 08 54 02 fd 5b d7 30 77 3d | 1....{f.T..[.0w=
	45 6c bb 37 ec d5 2e 9c 1e fe 20 7d ac 69 03 cb | El.7...... }.i..
	68 d7 d0 6c 27 17 4f 95 03 d0 54 c3 b7 e0 3c 7b | h..l'.O...T...<{
	7d 69 39 bb 77 0f fd 0e fe 87 bb 93 9e e8 4c 13 | }i9.w.........L.
	a2 60 1c a4 ec 99 64 62 fc 68 c6 42 ea 8a 5d 1a | .`....db.h.B..].
	4a 01 4f 0a ef be 7b 74 fb b5 f9 92 1a 90 c7 10 | J.O...{t........
	c8 ce a2 5b 87 43 6a 22 00 d4 9d 58 45 8f 23 84 | ...[.Cj"...XE.#.
	1b 9b f8 5b 44 ef e4 81 c5 75 f1 e2 1f 53 1a d0 | ...[D....u...S..
	c6 60 40 d4 1b 57 e6 fe b8 ff 79 21 37 41 b8 8e | .`@..W....y!7A..
	14 0a 8f 6e c6 67 68 27 e1 8a ed e5 ae 35 09 35 | ...n.gh'.....5.5
	da 56 68 56 0e 36 ee 47 86 26 42 66 00 17 00 0b | .VhV.6.G.&Bf....
	00 00                                           | ..
Response: 53
	80 02 00 00 00 35 00 00 00 00 00 00 00 22 00 20 | .....5.......". 
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	00 00 01 00 00                                  | .....
TPM2_RSA_Decrypt: 32
Command: 14
	80 01 00 00 00 0e 00 00 01 65 80 00 00 01       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
TPM2_FlushContext: Closed handle 0x80000001
RSA Encrypt Test Passed
Command: 102
	80 02 00 00 00 66 00 00 01 53 81 00 02 00 00 00 | .....f...S......
	00 1f 40 00 00 09 00 00 01 00 16 54 68 69 73 49 | ..@........ThisI
	73 4d 79 53 74 6f 72 61 67 65 4b 65 79 41 75 74 | sMyStorageKeyAut
	68 00 13 00 0f 54 68 69 73 49 73 4d 79 4b 65 79 | h....ThisIsMyKey
	41 75 74 68 00 00 00 18 00 23 00 0b 00 04 04 60 | Auth.....#.....`
	00 00 00 10 00 18 00 0b 00 03 00 10 00 00 00 00 | ................
	00 00 00 00 00 00                               | ......
Response: 428
	80 02 00 00 01 ac 00 00 00 00 00 00 01 99 00 7e | ...............~
	00 20 00 17 24 a8 de d8 94 89 cc d8 8c 90 53 b5 | . ..$.........S.
	8e 3d f6 29 3b e7 42 32 25 57 9b 0c ee 57 cc cf | .=.);.B2%W...W..
	ea 92 00 10 f4 47 59 95 6a 91 05 6d b0 d1 b2 8d | .....GY.j..m....
	7e 58 bd d6 0a 0b 34 42 4a 31 55 cb a0 25 1f 5f | ~X....4BJ1U..%._
	35 72 7a ce d0 f7 53 aa 8b 86 19 1c 1f a0 33 30 | 5rz...S.......30
	2a a5 a8 df f6 5d 09 3f 32 cb ff 8e e9 36 08 7a | *....].?2....6.z
	47 9c 0c 7f 69 04 e4 67 b6 55 19 c6 df 17 d7 33 | G...i..g.U.....3
	5e 3d 55 70 f4 e8 00 4b 4e 03 d5 41 31 c2 00 58 | ^=Up...KN..A1..X
	00 23 00 0b 00 04 04 60 00 00 00 10 00 18 00 0b | .#.....`........
	00 03 00 10 00 20 28 21 cb 59 ea 43 c6 53 e6 52 | ..... (!.Y.C.S.R
	4b a7 6c 14 a6 75 af d0 36 27 38 9f 6f 8b 95 09 | K.l..u..6'8.o...
	12 b3 fe 4f 8c 73 00 20 13 53 42 00 f3 26 17 52 | ...O.s. .SB..&.R
	eb e8 4f 05 64 d3 d9 a5 fa 1d 84 52 66 df fe a6 | ..O.d......Rf...
	7e 86 81 82 b3 09 0a 93 00 73 00 00 00 00 00 20 | ~........s..... 
	e3 b0 c4 42 98 fc 1c 14 9a fb f4 c8 99 6f b9 24 | ...B.........o.$
	27 ae 41 e4 64 9b 93 4c a4 95 99 1b 78 52 b8 55 | '.A.d..L....xR.U
	01 00 0b 00 22 00 0b 23 24 f6 f5 3c 45 70 7c 3e | ...."..#$..<Ep|>
	0a d7 78 3e bc 01 ae cb d9 73 0a 54 49 92 14 63 | ..x>.....s.TI..c
	05 c6 a8 16 ed fc fc 00 22 00 0b 46 d7 59 62 84 | ........"..F.Yb.
	e1 eb e4 2d 48 d7 0c 35 33 ae 63 81 23 f2 e4 36 | ...-H..53.c.#..6
	e0 da 08 00 f4 6f 96 91 76 99 c8 00 00 00 20 31 | .....o..v..... 1
	13 9d 8e 56 51 c7 7b 71 e0 0c 82 ea 60 5b ca e3 | ...VQ.{q....`[..
	eb 54 32 ad f6 e1 9b 94 bb fa c9 b0 b6 8c 59 80 | .T2...........Y.
	21 40 00 00 01 00 20 2a d3 9b b8 85 0e 63 96 55 | !@.... *.....c.U
	6d 45 74 41 cc 76 43 9b 91 db 7a df ac 1b e8 75 | mEtA.vC...z....u
	90 9d 30 34 d6 63 05 00 00 01 00 00             | ..04.c......
TPM2_Create key: pub 88, priv 126
Command: 267
	80 02 00 00 01 0b 00 00 01 57 81 00 02 00 00 00 | .........W......
	00 1f 40 00 00 09 00 00 01 00 16 54 68 69 73 49 | ..@........ThisI
	73 4d 79 53 74 6f 72 61 67 65 4b 65 79 41 75 74 | sMyStorageKeyAut
	68 00 7e 00 20 00 17 24 a8 de d8 94 89 cc d8 8c | h.~. ..$........
	90 53 b5 8e 3d f6 29 3b e7 42 32 25 57 9b 0c ee | .S..=.);.B2%W...
	57 cc cf ea 92 00 10 f4 47 59 95 6a 91 05 6d b0 | W.......GY.j..m.
	d1 b2 8d 7e 58 bd d6 0a 0b 34 42 4a 31 55 cb a0 | ...~X....4BJ1U..
	25 1f 5f 35 72 7a ce d0 f7 53 aa 8b 86 19 1c 1f | %._5rz...S......
	a0 33 30 2a a5 a8 df f6 5d 09 3f 32 cb ff 8e e9 | .30*....].?2....
	36 08 7a 47 9c 0c 7f 69 04 e4 67 b6 55 19 c6 df | 6.zG...i..g.U...
	17 d7 33 5e 3d 55 70 f4 e8 00 4b 4e 03 d5 41 31 | ..3^=Up...KN..A1
	c2 00 58 00 23 00 0b 00 04 04 60 00 00 00 10 00 | ..X.#.....`.....
	18 00 0b 00 03 00 10 00 20 28 21 cb 59 ea 43 c6 | ........ (!.Y.C.
	53 e6 52 4b a7 6c 14 a6 75 af d0 36 27 38 9f 6f | S.RK.l..u..6'8.o
	8b 95 09 12 b3 fe 4f 8c 73 00 20 13 53 42 00 f3 | ......O.s. .SB..
	26 17 52 eb e8 4f 05 64 d3 d9 a5 fa 1d 84 52 66 | &.R..O.d......Rf
	df fe a6 7e 86 81 82 b3 09 0a 93                | ...~.......
Response: 59
	80 02 00 00 00 3b 00 00 00 00 80 00 00 01 00 00 | .....;..........
	00 24 00 22 00 0b 2f a4 f8 c5 c7 27 76 e7 a9 ef | .$."../....'v...
	cc 0e 6d 81 05 ef 44 36 82 96 05 1a 68 24 c2 55 | ..m...D6....h$.U
	7e 14 ed 19 68 33 00 00 01 00 00                | ~...h3.....
TPM2_Load Key Handle 0x80000001
Command: 88
	80 02 00 00 00 58 00 00 01 5d 80 00 00 01 00 00 | .....X...]......
	00 18 40 00 00 09 00 00 01 00 0f 54 68 69 73 49 | ..@........ThisI
	73 4d 79 4b 65 79 41 75 74 68 00 20 11 11 11 11 | sMyKeyAuth. ....
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 00 18 00 0b | ................
	80 24 40 00 00 07 00 00                         | .$@.....
Response: 91
	80 02 00 00 00 5b 00 00 00 00 00 00 00 48 00 18 | .....[.......H..
	00 0b 00 20 08 39 23 4d a2 93 27 e0 67 e3 0a 04 | ... .9#M..'.g...
	4d 75 5c ae 24 b1 4c 55 ee 26 ce 42 4a ab 28 d8 | Mu\.$.LU.&.BJ.(.
	2c 67 d9 fb 00 20 7c 20 b3 41 4c 71 98 16 95 4d | ,g... | .ALq...M
	36 76 92 cf 03 d8 6e 71 5a 57 b6 a6 8f df 38 34 | 6v....nqZW....84
	00 b9 26 f4 a0 9e 00 00 01 00 00                | ..&........
TPM2_Sign: ECC R 32, S 32
Command: 120
	80 01 00 00 00 78 00 00 01 77 80 00 00 01 00 20 | .....x...w..... 
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	00 18 00 0b 00 20 08 39 23 4d a2 93 27 e0 67 e3 | ..... .9#M..'.g.
	0a 04 4d 75 5c ae 24 b1 4c 55 ee 26 ce 42 4a ab | ..Mu\.$.LU.&.BJ.
	28 d8 2c 67 d9 fb 00 20 7c 20 b3 41 4c 71 98 16 | (.,g... | .ALq..
	95 4d 36 76 92 cf 03 d8 6e 71 5a 57 b6 a6 8f df | .M6v....nqZW....
	38 34 00 b9 26 f4 a0 9e                         | 84..&...
Response: 50
	80 01 00 00 00 32 00 00 00 00 80 22 40 00 00 01 | .....2....."@...
	00 20 b8 e9 2b f8 8e 8c 20 6c b8 60 bb b9 f7 82 | . ..+... l.`....
	61 36 e9 0c 37 e2 9e bb af 06 20 05 a1 34 6c f1 | a6..7..... ..4l.
	18 b7                                           | ..
TPM2_VerifySignature: Tag 32802
Command: 14
	80 01 00 00 00 0e 00 00 01 65 80 00 00 01       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
TPM2_FlushContext: Closed handle 0x80000001
ECC Sign/Verify Passed
Command: 102
	80 02 00 00 00 66 00 00 01 53 81 00 02 00 00 00 | .....f...S......
	00 1f 40 00 00 09 00 00 01 00 16 54 68 69 73 49 | ..@........ThisI
	73 4d 79 53 74 6f 72 61 67 65 4b 65 79 41 75 74 | sMyStorageKeyAut
	68 00 13 00 0f 54 68 69 73 49 73 4d 79 4b 65 79 | h....ThisIsMyKey
	41 75 74 68 00 00 00 18 00 23 00 0b 00 02 04 60 | Auth.....#.....`
	00 00 00 10 00 19 00 0b 00 03 00 10 00 00 00 00 | ................
	00 00 00 00 00 00                               | ......
Response: 428
	80 02 00 00 01 ac 00 00 00 00 00 00 01 99 00 7e | ...............~
	00 20 c9 d7 10 6d 8a a4 95 77 d7 5f 18 7e bf 8c | . ...m...w._.~..
	8f f8 f6 7c 8f 87 70 32 22 78 3e 8d a0 a2 94 f4 | ...|..p2"x>.....
	22 03 00 10 4b ae 0a 0f a5 ec 33 36 18 fe 37 6c | "...K.....36..7l
	bc bb f5 86 46 1f db ec b0 85 49 cb c3 d6 07 1e | ....F.....I.....
	50 d6 b7 f2 00 90 cd f7 2c 41 cc 75 a5 7c 50 40 | P.......,A.u.|P@
	5e b5 c2 2d 20 1d 9f e4 0c ad a9 c3 77 c8 05 a5 | ^..- .......w...
	d8 a1 d9 a0 2b e8 e3 1b bc cd 54 0e f6 46 ef 26 | ....+.....T..F.&
	70 0f 8c e6 bd 76 92 19 a1 b6 72 3e ee cb 00 58 | p....v....r>...X
	00 23 00 0b 00 02 04 60 00 00 00 10 00 19 00 0b | .#.....`........
	00 03 00 10 00 20 f3 f0 4e 42 1c 0f b9 e5 91 6e | ..... ..NB.....n
	e0 98 be 2c 88 05 47 0a 36 44 40 d7 ff 73 34 e5 | ...,..G.6D@..s4.
	fd 5f 6e 04 f5 de 00 20 6d 04 6c 44 d2 98 ef 63 | ._n.... m.lD...c
	a1 9f f3 ff bb 05 a6 76 33 c6 7b 22 7f 2f 21 b6 | .......v3.{"./!.
	ea 39 44 18 a0 83 25 5c 00 73 00 00 00 00 00 20 | .9D...%\.s..... 
	e3 b0 c4 42 98 fc 1c 14 9a fb f4 c8 99 6f b9 24 | ...B.........o.$
	27 ae 41 e4 64 9b 93 4c a4 95 99 1b 78 52 b8 55 | '.A.d..L....xR.U
	01 00 0b 00 22 00 0b 23 24 f6 f5 3c 45 70 7c 3e | ...."..#$..<Ep|>
	0a d7 78 3e bc 01 ae cb d9 73 0a 54 49 92 14 63 | ..x>.....s.TI..c
	05 c6 a8 16 ed fc fc 00 22 00 0b 46 d7 59 62 84 | ........"..F.Yb.
	e1 eb e4 2d 48 d7 0c 35 33 ae 63 81 23 f2 e4 36 | ...-H..53.c.#..6
	e0 da 08 00 f4 6f 96 91 76 99 c8 00 00 00 20 31 | .....o..v..... 1
	13 9d 8e 56 51 c7 7b 71 e0 0c 82 ea 60 5b ca e3 | ...VQ.{q....`[..
	eb 54 32 ad f6 e1 9b 94 bb fa c9 b0 b6 8c 59 80 | .T2...........Y.
	21 40 00 00 01 00 20 c8 f0 7a 1f e1 99 37 64 80 | !@.... ..z...7d.
	d9 e0 2b 89 e6 22 cb e3 b3 5f a2 2e e0 36 73 27 | ..+.."..._...6s'
	58 15 ce 63 48 cb 8b 00 00 01 00 00             | X..cH.......
TPM2_Create key: pub 88, priv 126
Command: 267
	80 02 00 00 01 0b 00 00 01 57 81 00 02 00 00 00 | .........W......
	00 1f 40 00 00 09 00 00 01 00 16 54 68 69 73 49 | ..@........ThisI
	73 4d 79 53 74 6f 72 61 67 65 4b 65 79 41 75 74 | sMyStorageKeyAut
	68 00 7e 00 20 c9 d7 10 6d 8a a4 95 77 d7 5f 18 | h.~. ...m...w._.
	7e bf 8c 8f f8 f6 7c 8f 87 70 32 22 78 3e 8d a0 | ~.....|..p2"x>..
	a2 94 f4 22 03 00 10 4b ae 0a 0f a5 ec 33 36 18 | ..."...K.....36.
	fe 37 6c bc bb f5 86 46 1f db ec b0 85 49 cb c3 | .7l....F.....I..
	d6 07 1e 50 d6 b7 f2 00 90 cd f7 2c 41 cc 75 a5 | ...P.......,A.u.
	7c 50 40 5e b5 c2 2d 20 1d 9f e4 0c ad a9 c3 77 | |P@^..- .......w
	c8 05 a5 d8 a1 d9 a0 2b e8 e3 1b bc cd 54 0e f6 | .......+.....T..
	46 ef 26 70 0f 8c e6 bd 76 92 19 a1 b6 72 3e ee | F.&p....v....r>.
	cb 00 58 00 23 00 0b 00 02 04 60 00 00 00 10 00 | ..X.#.....`.....
	19 00 0b 00 03 00 10 00 20 f3 f0 4e 42 1c 0f b9 | ........ ..NB...
	e5 91 6e e0 98 be 2c 88 05 47 0a 36 44 40 d7 ff | ..n...,..G.6D@..
	73 34 e5 fd 5f 6e 04 f5 de 00 20 6d 04 6c 44 d2 | s4.._n.... m.lD.
	98 ef 63 a1 9f f3 ff bb 05 a6 76 33 c6 7b 22 7f | ..c.......v3.{".
	2f 21 b6 ea 39 44 18 a0 83 25 5c                | /!..9D...%\
Response: 59
	80 02 00 00 00 3b 00 00 00 00 80 00 00 01 00 00 | .....;..........
	00 24 00 22 00 0b e4 ac bc 9e 45 81 45 bb 26 d2 | .$."......E.E.&.
	e9 7e 94 5e a5 18 8a c2 6d 10 ec 8b 08 6b 5d ca | .~.^....m....k].
	cd be 23 b4 35 11 00 00 01 00 00                | ..#.5......
TPM2_Load Key Handle 0x80000001
Command: 14
	80 01 00 00 00 0e 00 00 01 63 80 00 00 01       | .........c....
Response: 150
	80 01 00 00 00 96 00 00 00 00 00 44 00 20 c8 63 | ...........D. .c
	e6 60 c3 4f 6c 39 38 ea 8c 0e 64 52 9a 05 ae 24 | .`.Ol98...dR...$
	8d 71 e2 ae e3 38 ea c7 51 60 77 51 7a bc 00 20 | .q...8..Q`wQz.. 
	41 d6 ce b3 f8 96 84 7a 33 e5 e3 c4 b3 24 a6 d3 | A......z3....$..
	a9 1c 91 d4 82 8d f6 9b 6b 02 58 ac e6 b7 70 5a | ........k.X...pZ
	00 44 00 20 f5 e1 e7 11 78 75 3f ba f1 ef 57 56 | .D. ....xu?...WV
	4f 18 5a d0 fa cd f3 a8 aa a9 82 f4 b1 55 4b 76 | O.Z..........UKv
	bb 29 4b 15 00 20 20 06 54 87 23 98 d3 e7 0b 26 | .)K..  .T.#....&
	10 50 42 66 c8 90 a8 f3 ae 8d c5 ed 43 31 26 4c | .PBf........C1&L
	36 e6 03 18 35 a5                               | 6...5.
TPM2_ECDH_KeyGen: zPt 68, pubPt 68
Command: 14
	80 01 00 00 00 0e 00 00 01 65 80 00 00 01       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
TPM2_FlushContext: Closed handle 0x80000001
ECC DH Generation Passed
Command: 45
	80 02 00 00 00 2d 00 00 01 2a 40 00 00 01 00 00 | .....-...*@.....
	00 09 40 00 00 09 00 00 01 00 00 00 00 00 0e 01 | ..@.............
	80 02 00 00 0b 02 06 00 06 00 00 04 00          | .............
Response: 19
	80 02 00 00 00 13 00 00 00 00 00 00 00 00 00 00 | ................
	01 00 00                                        | ...
TPM2_NV_DefineSpace: Auth 0x40000001, Idx 0x1800200, Attribs 0x33947654, Size 1024
Command: 803
	80 02 00 00 03 23 00 00 01 37 40 00 00 01 01 80 | .....#...7@.....
	02 00 00 00 00 09 40 00 00 09 00 00 01 00 00 03 | ......@.........
	00 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 00 00                                        | ...
Response: 19
	80 02 00 00 00 13 00 00 00 00 00 00 00 00 00 00 | ................
	01 00 00                                        | ...
TPM2_NV_Write: Auth 0x40000001, Idx 0x1800200, Offset 0, Size 768
Command: 291
	80 02 00 00 01 23 00 00 01 37 40 00 00 01 01 80 | .....#...7@.....
	02 00 00 00 00 09 40 00 00 09 00 00 01 00 00 01 | ......@.........
	00 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 03 00                                        | ...
Response: 19
	80 02 00 00 00 13 00 00 00 00 00 00 00 00 00 00 | ................
	01 00 00                                        | ...
TPM2_NV_Write: Auth 0x40000001, Idx 0x1800200, Offset 768, Size 256
Command: 35
	80 02 00 00 00 23 00 00 01 4e 40 00 00 01 01 80 | .....#...N@.....
	02 00 00 00 00 09 40 00 00 09 00 00 01 00 00 03 | ......@.........
	00 00 00                                        | ...
Response: 789
	80 02 00 00 03 15 00 00 00 00 00 00 03 02 03 00 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	00 00 01 00 00                                  | .....
TPM2_NV_Read: Auth 0x40000001, Idx 0x1800200, Offset 0, Size 768
Command: 35
	80 02 00 00 00 23 00 00 01 4e 40 00 00 01 01 80 | .....#...N@.....
	02 00 00 00 00 09 40 00 00 09 00 00 01 00 00 01 | ......@.........
	00 03 00                                        | ...
Response: 277
	80 02 00 00 01 15 00 00 00 00 00 00 01 02 01 00 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	00 00 01 00 00                                  | .....
TPM2_NV_Read: Auth 0x40000001, Idx 0x1800200, Offset 768, Size 256
Command: 14
	80 01 00 00 00 0e 00 00 01 69 01 80 02 00       | .........i....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 0e 01 80 02 00 | .....>..........
	00 0b 22 06 00 06 00 00 04 00 00 22 00 0b a5 89 | .."........"....
	bd 69 b5 f1 44 60 05 f3 f4 6c fa 07 99 b8 92 8a | .i..D`...l......
	1e a8 e6 29 a0 f2 67 22 80 00 f5 45 33 f1       | ...)..g"...E3.
TPM2_NV_ReadPublic: Sz 14, Idx 0x1800200, nameAlg 11, Attr 0x22060006, authPol 0, dataSz 1024, name 34
Command: 31
	80 02 00 00 00 1f 00 00 01 22 40 00 00 01 01 80 | ........."@.....
	02 00 00 00 00 09 40 00 00 09 00 00 01 00 00    | ......@........
Response: 19
	80 02 00 00 00 13 00 00 00 00 00 00 00 00 00 00 | ................
	01 00 00                                        | ...
TPM2_NV_UndefineSpace: Auth 0x40000001, Idx 0x1800200
NV Test on index 0x1800200 with 1024 bytes passed
Command: 12
	80 01 00 00 00 0c 00 00 01 45 00 00             | .........E..
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
```


### TPM Native Tests With Debug Enabled:

Built with `./configure --enable-debug` or `#define DEBUG_WOLFTPM`


```
./examples/native/native_test
TPM2 Demo using Native API's
TPM2: Caps 0x30000697, Did 0x001b, Vid 0x15d1, Rid 0x10 
Command: 12
	80 01 00 00 00 0c 00 00 01 44 00 00             | .........D..
Response: 10
	80 01 00 00 00 0a 00 00 01 00                   | ..........
TPM2_Startup pass
Command: 11
	80 01 00 00 00 0b 00 00 01 43 01                | .........C.
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
TPM2_SelfTest pass
Command: 10
	80 01 00 00 00 0a 00 00 01 7c                   | .........|
Response: 26
	80 01 00 00 00 1a 00 00 00 00 00 0a 00 01 f9 db | ................
	00 00 00 00 00 00 00 00 00 00                   | ..........
TPM2_GetTestResult: Size 10, Rc 0x0
	00 01 f9 db 00 00 00 00 00 00                   | ..........
Command: 16
	80 01 00 00 00 10 00 00 01 42 00 00 00 01 00 01 | .........B......
Response: 14
	80 01 00 00 00 0e 00 00 00 00 00 00 00 00       | ..............
TPM2_IncrementalSelfTest: Rc 0x0, Alg 0x1 (Todo 0)
Command: 22
	80 01 00 00 00 16 00 00 01 7a 00 00 00 06 00 00 | .........z......
	01 00 00 00 00 01                               | ......
Response: 27
	80 01 00 00 00 1b 00 00 00 00 01 00 00 00 06 00 | ................
	00 00 01 00 00 01 00 32 2e 30 00                | .......2.0.
TPM2_GetCapability: Property FamilyIndicator 0x322e3000
Command: 22
	80 01 00 00 00 16 00 00 01 7a 00 00 00 06 00 00 | .........z......
	01 12 00 00 00 01                               | ......
Response: 27
	80 01 00 00 00 1b 00 00 00 00 01 00 00 00 06 00 | ................
	00 00 01 00 00 01 12 00 00 00 18                | ...........
TPM2_GetCapability: Property PCR Count 24
Command: 12
	80 01 00 00 00 0c 00 00 01 7b 00 20             | .........{. 
Response: 44
	80 01 00 00 00 2c 00 00 00 00 00 20 fb 1c d0 d0 | .....,..... ....
	5d ee a2 49 f7 b2 5a 38 72 87 7b aa 10 e8 33 12 | ]..I..Z8r.{...3.
	dc b7 78 54 46 6f 61 d9 7d 1f b7 0e             | ..xTFoa.}...
TPM2_GetRandom: Got 32 bytes
	fb 1c d0 d0 5d ee a2 49 f7 b2 5a 38 72 87 7b aa | ....]..I..Z8r.{.
	10 e8 33 12 dc b7 78 54 46 6f 61 d9 7d 1f b7 0e | ..3...xTFoa.}...
Command: 44
	80 01 00 00 00 2c 00 00 01 46 00 20 fb 1c d0 d0 | .....,...F. ....
	5d ee a2 49 f7 b2 5a 38 72 87 7b aa 10 e8 33 12 | ]..I..Z8r.{...3.
	dc b7 78 54 46 6f 61 d9 7d 1f b7 0e             | ..xTFoa.}...
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
TPM2_StirRandom: success
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 01 00 00                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 01 00 00 00 00 00 01 00 20 de 96 | ............. ..
	1d 6b 9f 26 9c 61 ba 48 52 12 34 80 da ac ed 4c | .k.&.a.HR.4....L
	6a 5d 6d f1 90 94 1f b2 0b e4 17 d7 8a 2e       | j]m...........
TPM2_PCR_Read: Index 0, Digest Sz 32, Update Counter 22
	de 96 1d 6b 9f 26 9c 61 ba 48 52 12 34 80 da ac | ...k.&.a.HR.4...
	ed 4c 6a 5d 6d f1 90 94 1f b2 0b e4 17 d7 8a 2e | .Lj]m...........
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 02 00 00                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 02 00 00 00 00 00 01 00 20 00 00 | ............. ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 1, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 04 00 00                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 04 00 00 00 00 00 01 00 20 00 00 | ............. ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 2, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 08 00 00                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 08 00 00 00 00 00 01 00 20 00 00 | ............. ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 3, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 10 00 00                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 10 00 00 00 00 00 01 00 20 00 00 | ............. ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 4, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 20 00 00                                     | . ..
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 20 00 00 00 00 00 01 00 20 00 00 | ..... ....... ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 5, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 40 00 00                                     | .@..
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 40 00 00 00 00 00 01 00 20 00 00 | .....@....... ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 6, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 80 00 00                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 80 00 00 00 00 00 01 00 20 00 00 | ............. ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 7, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 01 00                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 01 00 00 00 00 01 00 20 00 00 | ............. ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 8, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 02 00                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 02 00 00 00 00 01 00 20 00 00 | ............. ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 9, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 04 00                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 04 00 00 00 00 01 00 20 00 00 | ............. ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 10, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 08 00                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 08 00 00 00 00 01 00 20 00 00 | ............. ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 11, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 10 00                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 10 00 00 00 00 01 00 20 00 00 | ............. ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 12, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 20 00                                     | .. .
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 20 00 00 00 00 01 00 20 00 00 | ...... ...... ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 13, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 40 00                                     | ..@.
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 40 00 00 00 00 01 00 20 00 00 | ......@...... ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 14, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 80 00                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 80 00 00 00 00 01 00 20 00 00 | ............. ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 15, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 00 01                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 00 01 00 00 00 01 00 20 00 00 | ............. ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 16, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 00 02                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 00 02 00 00 00 01 00 20 ff ff | ............. ..
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff       | ..............
TPM2_PCR_Read: Index 17, Digest Sz 32, Update Counter 22
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 00 04                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 00 04 00 00 00 01 00 20 ff ff | ............. ..
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff       | ..............
TPM2_PCR_Read: Index 18, Digest Sz 32, Update Counter 22
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 00 08                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 00 08 00 00 00 01 00 20 ff ff | ............. ..
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff       | ..............
TPM2_PCR_Read: Index 19, Digest Sz 32, Update Counter 22
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 00 10                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 00 10 00 00 00 01 00 20 ff ff | ............. ..
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff       | ..............
TPM2_PCR_Read: Index 20, Digest Sz 32, Update Counter 22
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 00 20                                     | ... 
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 00 20 00 00 00 01 00 20 ff ff | ....... ..... ..
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff       | ..............
TPM2_PCR_Read: Index 21, Digest Sz 32, Update Counter 22
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 00 40                                     | ...@
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 00 40 00 00 00 01 00 20 ff ff | .......@..... ..
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff       | ..............
TPM2_PCR_Read: Index 22, Digest Sz 32, Update Counter 22
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
	ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 00 00 80                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 16 00 00 | .....>..........
	00 01 00 0b 03 00 00 80 00 00 00 01 00 20 00 00 | ............. ..
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00       | ..............
TPM2_PCR_Read: Index 23, Digest Sz 32, Update Counter 22
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 65
	80 02 00 00 00 41 00 00 01 82 00 00 00 00 00 00 | .....A..........
	00 09 40 00 00 09 00 00 01 00 00 00 00 00 01 00 | ..@.............
	0b 00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e | ................
	0f 10 11 12 13 14 15 16 17 18 19 1a 1b 1c 1d 1e | ................
	1f                                              | .
Response: 19
	80 02 00 00 00 13 00 00 00 00 00 00 00 00 00 00 | ................
	01 00 00                                        | ...
TPM2_PCR_Extend success
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 0b | .........~......
	03 01 00 00                                     | ....
Response: 62
	80 01 00 00 00 3e 00 00 00 00 00 00 00 17 00 00 | .....>..........
	00 01 00 0b 03 01 00 00 00 00 00 01 00 20 8e fc | ............. ..
	5b 47 15 71 24 25 df 43 01 da e7 99 60 40 e0 6c | [G.q$%.C....`@.l
	6f a0 be 40 38 81 8d ef d4 ce 7b c6 c2 80       | o..@8.....{...
TPM2_PCR_Read: Index 0, Digest Sz 32, Update Counter 23
	8e fc 5b 47 15 71 24 25 df 43 01 da e7 99 60 40 | ..[G.q$%.C....`@
	e0 6c 6f a0 be 40 38 81 8d ef d4 ce 7b c6 c2 80 | .lo..@8.....{...
Command: 59
	80 01 00 00 00 3b 00 00 01 76 40 00 00 07 40 00 | .....;...v@...@.
	00 07 00 20 12 56 de e7 dd df ab b3 01 95 50 de | ... .V........P.
	d6 5f 26 fd c9 93 f3 d0 96 c6 67 45 3c 60 54 e9 | ._&.......gE<`T.
	a9 0b b7 cc 00 00 01 00 10 00 0b                | ...........
Response: 48
	80 01 00 00 00 30 00 00 00 00 03 00 00 00 00 20 | .....0......... 
	c7 76 20 1b f9 c9 2b 99 78 d3 fa 26 8b f5 8e bf | .v ...+.x..&....
	8e 38 1e 5e 49 a4 df 7d b0 01 f7 3e b7 11 60 14 | .8.^I..}...>..`.
TPM2_StartAuthSession: sessionHandle 0x3000000
Command: 14
	80 01 00 00 00 0e 00 00 01 89 03 00 00 00       | ..............
Response: 44
	80 01 00 00 00 2c 00 00 00 00 00 20 00 00 00 00 | .....,..... ....
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00             | ............
TPM2_PolicyGetDigest: size 32
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
Command: 20
	80 01 00 00 00 14 00 00 01 7e 00 00 00 01 00 04 | .........~......
	03 01 00 00                                     | ....
Response: 50
	80 01 00 00 00 32 00 00 00 00 00 00 00 17 00 00 | .....2..........
	00 01 00 04 03 01 00 00 00 00 00 01 00 14 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00                                           | ..
TPM2_PCR_Read: Index 0, Digest Sz 20, Update Counter 23
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00                                     | ....
wc_Hash of PCR[0]: size 32
	66 68 7a ad f8 62 bd 77 6c 8f c1 8b 8e 9f 8e 20 | fhz..b.wl...... 
	08 97 14 85 6e e2 33 b3 90 2a 59 1d 0d 5f 29 25 | ....n.3..*Y.._)%
Command: 58
	80 01 00 00 00 3a 00 00 01 7f 03 00 00 00 00 20 | .....:......... 
	66 68 7a ad f8 62 bd 77 6c 8f c1 8b 8e 9f 8e 20 | fhz..b.wl...... 
	08 97 14 85 6e e2 33 b3 90 2a 59 1d 0d 5f 29 25 | ....n.3..*Y.._)%
	00 00 00 01 00 04 03 01 00 00                   | ..........
Response: 10
	80 01 00 00 00 0a 00 00 01 c4                   | ..........
TPM2_PolicyPCR failed 0x1c4: TPM_RC_AUTHSIZE: The value of authorizationSize is out of range or the number of octets in the Authorization Area is greater than required
TPM2_PolicyPCR: Updated
Command: 14
	80 01 00 00 00 0e 00 00 01 80 03 00 00 00       | ..............
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
TPM2_PolicyRestart: Done
Command: 36
	80 01 00 00 00 24 00 00 01 86 00 16 54 68 69 73 | .....$......This
	49 73 41 53 65 63 72 65 74 55 73 61 67 65 41 75 | IsASecretUsageAu
	74 68 00 0b                                     | th..
Response: 14
	80 01 00 00 00 0e 00 00 00 00 80 00 00 00       | ..............
TPM2_HashSequenceStart: sequenceHandle 0x80000000
Command: 107
	80 02 00 00 00 6b 00 00 01 5c 80 00 00 00 00 00 | .....k...\......
	00 1f 40 00 00 09 00 00 01 00 16 54 68 69 73 49 | ..@........ThisI
	73 41 53 65 63 72 65 74 55 73 61 67 65 41 75 74 | sASecretUsageAut
	68 00 38 61 62 63 64 62 63 64 65 63 64 65 66 64 | h.8abcdbcdecdefd
	65 66 67 65 66 67 68 66 67 68 69 67 68 69 6a 68 | efgefghfghighijh
	69 6a 6b 69 6a 6b 6c 6a 6b 6c 6d 6b 6c 6d 6e 6c | ijkijkljklmklmnl
	6d 6e 6f 6d 6e 6f 70 6e 6f 70 71                | mnomnopnopq
Response: 19
	80 02 00 00 00 13 00 00 00 00 00 00 00 00 00 00 | ................
	01 00 00                                        | ...
Command: 55
	80 02 00 00 00 37 00 00 01 3e 80 00 00 00 00 00 | .....7...>......
	00 1f 40 00 00 09 00 00 01 00 16 54 68 69 73 49 | ..@........ThisI
	73 41 53 65 63 72 65 74 55 73 61 67 65 41 75 74 | sASecretUsageAut
	68 00 00 40 00 00 07                            | h..@...
Response: 61
	80 02 00 00 00 3d 00 00 00 00 00 00 00 2a 00 20 | .....=.......*. 
	24 8d 6a 61 d2 06 38 b8 e5 c0 26 93 0c 3e 60 39 | $.ja..8...&..>`9
	a3 3c e4 59 64 ff 21 67 f6 ec ed d4 19 db 06 c1 | .<.Yd.!g........
	80 24 40 00 00 07 00 00 00 00 01 00 00          | .$@..........
Hash SHA256 test success
Command: 355
	80 02 00 00 01 63 00 00 01 31 40 00 00 0b 00 00 | .....c...1@.....
	00 09 40 00 00 09 00 00 01 00 00 00 04 00 00 00 | ..@.............
	00 01 3a 00 01 00 0b 00 03 00 b2 00 20 83 71 97 | ..:......... .q.
	67 44 84 b3 f8 1a 90 cc 8d 46 a5 d7 24 fd 52 d7 | gD.......F..$.R.
	6e 06 52 0b 64 f2 a1 da 1b 33 14 69 aa 00 06 00 | n.R.d....3.i....
	80 00 43 00 10 08 00 00 00 00 00 01 00 00 00 00 | ..C.............
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00                                        | ...
Response: 506
	80 02 00 00 01 fa 00 00 00 00 80 00 00 00 00 00 | ................
	01 e3 01 3a 00 01 00 0b 00 03 00 b2 00 20 83 71 | ...:......... .q
	97 67 44 84 b3 f8 1a 90 cc 8d 46 a5 d7 24 fd 52 | .gD.......F..$.R
	d7 6e 06 52 0b 64 f2 a1 da 1b 33 14 69 aa 00 06 | .n.R.d....3.i...
	00 80 00 43 00 10 08 00 00 00 00 00 01 00 e5 1f | ...C............
	d6 59 c3 78 58 5a 13 06 ef b3 05 b9 f5 dc f0 9c | .Y.xXZ..........
	37 aa b5 d1 db 29 6c 4d 4e 13 48 ce 6d eb a3 81 | 7....)lMN.H.m...
	16 eb f5 b5 7b 49 89 b7 2d 5f 7f d5 15 9a ce eb | ....{I..-_......
	9f 63 8d fd 77 b4 cb b7 1c 7d 51 d8 de c7 a6 eb | .c..w....}Q.....
	8a 09 3a 9f 5e c9 a6 1a 40 86 89 6d 99 a7 82 97 | ..:.^...@..m....
	f8 e3 7b 2d 8a de ba d9 59 2f c1 9b 8b ee dc e8 | ..{-....Y/......
	7c 07 db 57 22 84 15 bd 05 8c 4f 1d c7 ae fe 50 | |..W".....O....P
	9d 4a eb 1c be 7a 13 c6 57 34 7c 8c 39 da 7b 8a | .J...z..W4|.9.{.
	2e 32 e5 0c 44 6c 86 4e 02 a6 03 9a f9 49 ce 07 | .2..Dl.N.....I..
	29 e6 35 bf eb 2b f2 e0 23 ed c6 06 b1 eb 8d df | ).5..+..#.......
	d8 e8 9a b8 63 3b a3 ba 91 eb 1a 9f 99 0b 17 ba | ....c;..........
	4c 54 88 26 ca c4 3a b6 52 af c8 04 e1 61 86 8f | LT.&..:.R....a..
	34 ad 43 f3 6d 3c 0d 84 65 dc eb f5 21 73 10 c0 | 4.C.m<..e...!s..
	75 0a 1a 0c e9 1d 3a b3 5a 9b c2 92 f3 39 d7 d7 | u.....:.Z....9..
	3a 6b 6c 53 79 40 29 b4 21 47 b7 6d 92 71 b8 5f | :klSy@).!G.m.q._
	b2 11 52 c4 4a e9 04 14 7f c6 0d 77 43 b9 00 37 | ..R.J......wC..7
	00 00 00 00 00 20 e3 b0 c4 42 98 fc 1c 14 9a fb | ..... ...B......
	f4 c8 99 6f b9 24 27 ae 41 e4 64 9b 93 4c a4 95 | ...o.$'.A.d..L..
	99 1b 78 52 b8 55 01 00 10 00 04 40 00 00 0b 00 | ..xR.U.....@....
	04 40 00 00 0b 00 00 00 20 28 d0 26 fa fd 74 91 | .@...... (.&..t.
	06 74 3e 27 c4 28 05 51 58 5e 5d 17 66 8e b5 21 | .t>'.(.QX^].f..!
	83 5e d6 01 27 ef fc 05 d4 80 21 40 00 00 0b 00 | .^..'.....!@....
	20 8e cd d1 0d 44 5e 36 5d be 7a 3d f7 3a 80 c0 |  ....D^6].z=.:..
	27 1d 4f 83 12 ae ef 25 f7 b7 bb a1 5c e6 bb 8d | '.O....%....\...
	29 00 22 00 0b 99 da 0a bc 8f c1 fc f9 3f 0f a2 | )."..........?..
	bf ca 2c c2 8d 11 02 be d4 be 5f 77 8c c1 a7 8f | ..,......._w....
	f0 59 25 18 8d 00 00 01 00 00                   | .Y%.......
TPM2_CreatePrimary: Endorsement 0x80000000 (314 bytes)
Command: 338
	80 02 00 00 01 52 00 00 01 31 40 00 00 01 00 00 | .....R...1@.....
	00 09 40 00 00 09 00 00 01 00 00 00 13 00 0f 57 | ..@............W
	6f 6c 66 54 50 4d 50 61 73 73 77 6f 72 64 00 00 | olfTPMPassword..
	01 1a 00 01 00 0b 00 03 04 72 00 00 00 06 00 80 | .........r......
	00 43 00 10 08 00 00 00 00 00 01 00 00 00 00 00 | .C..............
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00                                           | ..
Response: 474
	80 02 00 00 01 da 00 00 00 00 80 00 00 01 00 00 | ................
	01 c3 01 1a 00 01 00 0b 00 03 04 72 00 00 00 06 | ...........r....
	00 80 00 43 00 10 08 00 00 00 00 00 01 00 af 17 | ...C............
	14 f3 dd 71 f5 4b ce 09 04 40 18 30 25 18 97 4e | ...q.K...@.0%..N
	7d 97 e2 6a 99 7f 1c 79 d8 f1 bc eb 97 f1 6a 63 | }..j...y......jc
	6d 43 60 a1 30 5a fc 14 b5 e3 d1 e0 b7 39 90 43 | mC`.0Z.......9.C
	30 11 8c e7 01 09 e8 01 a8 bd e9 60 08 6d 8c a3 | 0..........`.m..
	c1 a0 a5 40 e6 56 dc 98 84 75 4a ca 69 17 d9 0e | ...@.V...uJ.i...
	f7 66 70 ce 64 51 20 b8 70 c0 d5 86 ad b4 81 ab | .fp.dQ .p.......
	bd f2 43 73 3f 8c 2a 39 f1 3a 21 18 82 c3 1d d3 | ..Cs?.*9.:!.....
	39 d6 73 84 51 90 d3 f0 7a 08 cc a7 f4 1c 6a 6f | 9.s.Q...z.....jo
	27 48 43 bb ed af 3c a7 9c 6e da b7 12 04 28 14 | 'HC...<..n....(.
	07 23 72 c3 01 e6 c6 c9 b6 ff 86 3d c2 4e dc 7c | .#r........=.N.|
	66 1f 62 0b 88 32 26 19 0c cd 72 63 9c aa 39 ef | f.b..2&...rc..9.
	87 6b d1 2f 2f 3a 03 96 7e 34 b9 06 1a 4f 6e f2 | .k.//:..~4...On.
	f3 0f 9f 4c 33 37 35 e0 93 f6 be 31 5c 6f b3 83 | ...L375....1\o..
	50 88 57 71 31 9f 1d 57 fb 9b 4b 8d 5d c1 66 b1 | P.Wq1..W..K.].f.
	be ea f6 5e 00 15 91 13 76 ab c8 6b e4 ad 86 a9 | ...^....v..k....
	57 ad fb e2 2c 45 c5 90 11 cc 6f bd 5e f3 00 37 | W...,E....o.^..7
	00 00 00 00 00 20 e3 b0 c4 42 98 fc 1c 14 9a fb | ..... ...B......
	f4 c8 99 6f b9 24 27 ae 41 e4 64 9b 93 4c a4 95 | ...o.$'.A.d..L..
	99 1b 78 52 b8 55 01 00 10 00 04 40 00 00 01 00 | ..xR.U.....@....
	04 40 00 00 01 00 00 00 20 5d a0 41 ba c0 ee 31 | .@...... ].A...1
	35 ae bb 0c ad fb a4 97 c6 a1 87 7f ae 83 2d d3 | 5.............-.
	d1 f8 f7 a8 71 b8 25 e8 54 80 21 40 00 00 01 00 | ....q.%.T.!@....
	20 18 d0 cd bd 8a c2 3f f4 4e 82 ef 4b e0 ea 16 |  ......?.N..K...
	4a c6 4a 8c 23 de a7 7e d3 c3 77 91 e0 13 6d ea | J.J.#..~..w...m.
	3e 00 22 00 0b 23 24 f6 f5 3c 45 70 7c 3e 0a d7 | >."..#$..<Ep|>..
	78 3e bc 01 ae cb d9 73 0a 54 49 92 14 63 05 c6 | x>.....s.TI..c..
	a8 16 ed fc fc 00 00 01 00 00                   | ..........
TPM2_CreatePrimary: Storage 0x80000001 (282 bytes)
Command: 332
	80 01 00 00 01 4c 00 00 01 67 00 00 01 3a 00 01 | .....L...g...:..
	00 0b 00 03 00 b2 00 20 83 71 97 67 44 84 b3 f8 | ....... .q.gD...
	1a 90 cc 8d 46 a5 d7 24 fd 52 d7 6e 06 52 0b 64 | ....F..$.R.n.R.d
	f2 a1 da 1b 33 14 69 aa 00 06 00 80 00 43 00 10 | ....3.i......C..
	08 00 00 00 00 00 01 00 e5 1f d6 59 c3 78 58 5a | ...........Y.xXZ
	13 06 ef b3 05 b9 f5 dc f0 9c 37 aa b5 d1 db 29 | ..........7....)
	6c 4d 4e 13 48 ce 6d eb a3 81 16 eb f5 b5 7b 49 | lMN.H.m.......{I
	89 b7 2d 5f 7f d5 15 9a ce eb 9f 63 8d fd 77 b4 | ..-_.......c..w.
	cb b7 1c 7d 51 d8 de c7 a6 eb 8a 09 3a 9f 5e c9 | ...}Q.......:.^.
	a6 1a 40 86 89 6d 99 a7 82 97 f8 e3 7b 2d 8a de | ..@..m......{-..
	ba d9 59 2f c1 9b 8b ee dc e8 7c 07 db 57 22 84 | ..Y/......|..W".
	15 bd 05 8c 4f 1d c7 ae fe 50 9d 4a eb 1c be 7a | ....O....P.J...z
	13 c6 57 34 7c 8c 39 da 7b 8a 2e 32 e5 0c 44 6c | ..W4|.9.{..2..Dl
	86 4e 02 a6 03 9a f9 49 ce 07 29 e6 35 bf eb 2b | .N.....I..).5..+
	f2 e0 23 ed c6 06 b1 eb 8d df d8 e8 9a b8 63 3b | ..#...........c;
	a3 ba 91 eb 1a 9f 99 0b 17 ba 4c 54 88 26 ca c4 | ..........LT.&..
	3a b6 52 af c8 04 e1 61 86 8f 34 ad 43 f3 6d 3c | :.R....a..4.C.m<
	0d 84 65 dc eb f5 21 73 10 c0 75 0a 1a 0c e9 1d | ..e...!s..u.....
	3a b3 5a 9b c2 92 f3 39 d7 d7 3a 6b 6c 53 79 40 | :.Z....9..:klSy@
	29 b4 21 47 b7 6d 92 71 b8 5f b2 11 52 c4 4a e9 | ).!G.m.q._..R.J.
	04 14 7f c6 0d 77 43 b9 40 00 00 07             | .....wC.@...
Response: 50
	80 01 00 00 00 32 00 00 00 00 80 00 00 02 00 22 | .....2........."
	00 0b 99 da 0a bc 8f c1 fc f9 3f 0f a2 bf ca 2c | ..........?....,
	c2 8d 11 02 be d4 be 5f 77 8c c1 a7 8f f0 59 25 | ......._w.....Y%
	18 8d                                           | ..
TPM2_LoadExternal: 0x80000002
Command: 84
	80 01 00 00 00 54 00 00 01 68 80 00 00 02 00 20 | .....T...h..... 
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	00 22 00 0b 99 da 0a bc 8f c1 fc f9 3f 0f a2 bf | ."..........?...
	ca 2c c2 8d 11 02 be d4 be 5f 77 8c c1 a7 8f f0 | .,......._w.....
	59 25 18 8d                                     | Y%..
Response: 338
	80 01 00 00 01 52 00 00 00 00 00 44 00 20 bc 7b | .....R.....D. .{
	b4 98 48 cf d5 25 de 74 e4 dc 93 98 96 01 d8 25 | ..H..%.t.......%
	40 20 e7 c7 cd ec 20 5b 36 8f 3e 1d e6 9b eb 5f | @ .... [6.>...._
	fe d9 c5 b6 53 70 a5 62 b6 65 81 07 1e a1 d4 d5 | ....Sp.b.e......
	e6 d1 93 3c f9 6d 56 96 0d f4 2d 25 d5 4a 15 92 | ...<.mV...-%.J..
	01 00 39 db fc c7 40 af 66 0e 7f bc 44 e9 58 79 | ..9...@.f...D.Xy
	2d ab 52 52 be 01 74 94 4d 66 99 3c 60 bb 03 cf | -.RR..t.Mf.<`...
	4f 7e 8d c7 19 4b e9 5d 93 4e df e0 36 2c d1 e1 | O~...K.].N..6,..
	84 67 7c 01 9f c2 a1 ed 76 81 dc b7 8e 8f 7b 5c | .g|.....v.....{\
	be 67 19 6e 10 67 ef fd 0f eb 42 dc 8a 75 27 37 | .g.n.g....B..u'7
	4e 02 d1 36 a6 cf 09 87 27 0b 51 f1 c7 07 62 8b | N..6....'.Q...b.
	f2 0f 4d e4 a1 89 e0 bf 4f 82 cd 9c 21 23 e7 c1 | ..M.....O...!#..
	c6 22 ab 80 74 f4 f1 3c d2 43 23 b4 dd 65 6b da | ."..t..<.C#..ek.
	96 43 cd 5b 4a a5 d8 f1 9e 9c 8d ef dc 48 49 f6 | .C.[J........HI.
	83 2a 3f 4f 7c d9 8c 06 b2 e3 53 da 18 ed 3b c7 | .*?O|.....S...;.
	48 dd b1 53 62 92 30 17 54 0a d5 27 02 f9 ba e3 | H..Sb.0.T..'....
	44 5c 0d 84 a9 55 a9 0c c6 6c 6c 86 9d 5e f3 7c | D\...U...ll..^.|
	f8 67 87 2b 95 14 1e 50 4c b5 ab c5 45 9d d1 2b | .g.+...PL...E..+
	c8 09 a0 3e fd a4 a7 b1 1e 5b 71 a1 84 f7 d6 56 | ...>.....[q....V
	a0 7b c2 aa 51 30 4d 57 ee da 86 a9 59 4e 38 1a | .{..Q0MW....YN8.
	e4 5e e7 90 51 9d 70 10 32 94 77 02 6a f8 d9 b7 | .^..Q.p.2.w.j...
	79 4f                                           | yO
TPM2_MakeCredential: credentialBlob 68, secret 256
Command: 14
	80 01 00 00 00 0e 00 00 01 73 80 00 00 02       | .........s....
Response: 398
	80 01 00 00 01 8e 00 00 00 00 01 3a 00 01 00 0b | ...........:....
	00 03 00 b2 00 20 83 71 97 67 44 84 b3 f8 1a 90 | ..... .q.gD.....
	cc 8d 46 a5 d7 24 fd 52 d7 6e 06 52 0b 64 f2 a1 | ..F..$.R.n.R.d..
	da 1b 33 14 69 aa 00 06 00 80 00 43 00 10 08 00 | ..3.i......C....
	00 00 00 00 01 00 e5 1f d6 59 c3 78 58 5a 13 06 | .........Y.xXZ..
	ef b3 05 b9 f5 dc f0 9c 37 aa b5 d1 db 29 6c 4d | ........7....)lM
	4e 13 48 ce 6d eb a3 81 16 eb f5 b5 7b 49 89 b7 | N.H.m.......{I..
	2d 5f 7f d5 15 9a ce eb 9f 63 8d fd 77 b4 cb b7 | -_.......c..w...
	1c 7d 51 d8 de c7 a6 eb 8a 09 3a 9f 5e c9 a6 1a | .}Q.......:.^...
	40 86 89 6d 99 a7 82 97 f8 e3 7b 2d 8a de ba d9 | @..m......{-....
	59 2f c1 9b 8b ee dc e8 7c 07 db 57 22 84 15 bd | Y/......|..W"...
	05 8c 4f 1d c7 ae fe 50 9d 4a eb 1c be 7a 13 c6 | ..O....P.J...z..
	57 34 7c 8c 39 da 7b 8a 2e 32 e5 0c 44 6c 86 4e | W4|.9.{..2..Dl.N
	02 a6 03 9a f9 49 ce 07 29 e6 35 bf eb 2b f2 e0 | .....I..).5..+..
	23 ed c6 06 b1 eb 8d df d8 e8 9a b8 63 3b a3 ba | #...........c;..
	91 eb 1a 9f 99 0b 17 ba 4c 54 88 26 ca c4 3a b6 | ........LT.&..:.
	52 af c8 04 e1 61 86 8f 34 ad 43 f3 6d 3c 0d 84 | R....a..4.C.m<..
	65 dc eb f5 21 73 10 c0 75 0a 1a 0c e9 1d 3a b3 | e...!s..u.....:.
	5a 9b c2 92 f3 39 d7 d7 3a 6b 6c 53 79 40 29 b4 | Z....9..:klSy@).
	21 47 b7 6d 92 71 b8 5f b2 11 52 c4 4a e9 04 14 | !G.m.q._..R.J...
	7f c6 0d 77 43 b9 00 22 00 0b 99 da 0a bc 8f c1 | ...wC.."........
	fc f9 3f 0f a2 bf ca 2c c2 8d 11 02 be d4 be 5f | ..?....,......._
	77 8c c1 a7 8f f0 59 25 18 8d 00 22 00 0b 99 da | w.....Y%..."....
	0a bc 8f c1 fc f9 3f 0f a2 bf ca 2c c2 8d 11 02 | ......?....,....
	be d4 be 5f 77 8c c1 a7 8f f0 59 25 18 8d       | ..._w.....Y%..
TPM2_ReadPublic Handle 0x80000002: pub 314, name 34, qualifiedName 34
Command: 14
	80 01 00 00 00 0e 00 00 01 65 80 00 00 02       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
Command: 109
	80 02 00 00 00 6d 00 00 01 53 80 00 00 01 00 00 | .....m...S......
	00 18 40 00 00 09 00 00 01 00 0f 57 6f 6c 66 54 | ..@........WolfT
	50 4d 50 61 73 73 77 6f 72 64 00 29 00 16 54 68 | PMPassword.)..Th
	69 73 49 73 41 53 65 63 72 65 74 55 73 61 67 65 | isIsASecretUsage
	41 75 74 68 00 0f 54 68 69 73 49 73 4d 79 48 6d | Auth..ThisIsMyHm
	61 63 4b 65 79 00 10 00 08 00 0b 00 04 04 40 00 | acKey.........@.
	00 00 05 00 0b 00 00 00 00 00 00 00 00          | .............
Response: 403
	80 02 00 00 01 93 00 00 00 00 00 00 01 80 00 8d | ................
	00 20 f3 48 ea 67 79 c3 f4 7b 5e db 6f 49 42 ef | . .H.gy..{^.oIB.
	a6 95 e5 b5 30 38 2c 85 c4 b2 98 ac a9 e9 ea 90 | ....08,.........
	9e 3f 00 10 f4 b0 8c 46 de 9c 20 b0 ea 22 91 39 | .?.....F.. ..".9
	32 76 48 b7 e0 ad 9b 14 a0 c0 b6 2b 96 ff 3d 0a | 2vH........+..=.
	5e cb 2b ee a1 51 6b 2e 5c f0 39 21 16 80 cb 52 | ^.+..Qk.\.9!...R
	10 e2 e8 8b 1f 82 28 8d ed 4f f5 32 a1 79 4f 4c | ......(..O.2.yOL
	af 07 e2 d6 aa b8 ec 31 c0 9e 4f 54 ff f3 be 13 | .......1..OT....
	9c 18 0c 19 b3 78 6e 48 05 bc 29 76 af de b6 fc | .....xnH..)v....
	60 c4 42 b3 4f 67 fa b9 21 d3 3f ca 86 00 30 00 | `.B.Og..!.?...0.
	08 00 0b 00 04 04 40 00 00 00 05 00 0b 00 20 b8 | ......@....... .
	5f 48 35 26 57 b0 cd 5b e7 83 dd 4d 25 7c 78 22 | _H5&W..[...M%|x"
	bc 75 5e 07 ce a0 c6 e2 a4 f9 74 9e f6 0e 0f 00 | .u^.......t.....
	73 00 00 00 00 00 20 e3 b0 c4 42 98 fc 1c 14 9a | s..... ...B.....
	fb f4 c8 99 6f b9 24 27 ae 41 e4 64 9b 93 4c a4 | ....o.$'.A.d..L.
	95 99 1b 78 52 b8 55 01 00 0b 00 22 00 0b 23 24 | ...xR.U...."..#$
	f6 f5 3c 45 70 7c 3e 0a d7 78 3e bc 01 ae cb d9 | ..<Ep|>..x>.....
	73 0a 54 49 92 14 63 05 c6 a8 16 ed fc fc 00 22 | s.TI..c........"
	00 0b 46 d7 59 62 84 e1 eb e4 2d 48 d7 0c 35 33 | ..F.Yb....-H..53
	ae 63 81 23 f2 e4 36 e0 da 08 00 f4 6f 96 91 76 | .c.#..6.....o..v
	99 c8 00 00 00 20 31 13 9d 8e 56 51 c7 7b 71 e0 | ..... 1...VQ.{q.
	0c 82 ea 60 5b ca e3 eb 54 32 ad f6 e1 9b 94 bb | ...`[...T2......
	fa c9 b0 b6 8c 59 80 21 40 00 00 01 00 20 d1 f6 | .....Y.!@.... ..
	7b 94 b4 39 9c f1 39 a0 e4 2c fb f6 c0 cd fc bb | {..9..9..,......
	ff 3f 1a 43 39 83 05 c8 c2 2c 5c d9 0f b4 00 00 | .?.C9....,\.....
	01 00 00                                        | ...
Create HMAC-SHA256 Key success, public 48, Private 141
Command: 235
	80 02 00 00 00 eb 00 00 01 57 80 00 00 01 00 00 | .........W......
	00 18 40 00 00 09 00 00 01 00 0f 57 6f 6c 66 54 | ..@........WolfT
	50 4d 50 61 73 73 77 6f 72 64 00 8d 00 20 f3 48 | PMPassword... .H
	ea 67 79 c3 f4 7b 5e db 6f 49 42 ef a6 95 e5 b5 | .gy..{^.oIB.....
	30 38 2c 85 c4 b2 98 ac a9 e9 ea 90 9e 3f 00 10 | 08,..........?..
	f4 b0 8c 46 de 9c 20 b0 ea 22 91 39 32 76 48 b7 | ...F.. ..".92vH.
	e0 ad 9b 14 a0 c0 b6 2b 96 ff 3d 0a 5e cb 2b ee | .......+..=.^.+.
	a1 51 6b 2e 5c f0 39 21 16 80 cb 52 10 e2 e8 8b | .Qk.\.9!...R....
	1f 82 28 8d ed 4f f5 32 a1 79 4f 4c af 07 e2 d6 | ..(..O.2.yOL....
	aa b8 ec 31 c0 9e 4f 54 ff f3 be 13 9c 18 0c 19 | ...1..OT........
	b3 78 6e 48 05 bc 29 76 af de b6 fc 60 c4 42 b3 | .xnH..)v....`.B.
	4f 67 fa b9 21 d3 3f ca 86 00 30 00 08 00 0b 00 | Og..!.?...0.....
	04 04 40 00 00 00 05 00 0b 00 20 b8 5f 48 35 26 | ..@....... ._H5&
	57 b0 cd 5b e7 83 dd 4d 25 7c 78 22 bc 75 5e 07 | W..[...M%|x".u^.
	ce a0 c6 e2 a4 f9 74 9e f6 0e 0f                | ......t....
Response: 59
	80 02 00 00 00 3b 00 00 00 00 80 00 00 02 00 00 | .....;..........
	00 24 00 22 00 0b e4 18 3e 1a da e5 20 c3 9e 65 | .$."....>... ..e
	6f f4 6c 2b 60 67 22 73 9b a7 fd 2d 24 12 48 36 | o.l+`g"s...-$.H6
	11 9a 7b 41 8d 1a 00 00 01 00 00                | ..{A.......
TPM2_Load New HMAC Key Handle 0x80000002
Command: 18
	80 01 00 00 00 12 00 00 01 6c 03 00 00 00 00 00 | .........l......
	01 50                                           | .P
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
TPM2_PolicyCommandCode: success
Command: 80
	80 02 00 00 00 50 00 00 01 50 80 00 00 02 80 00 | .....P...P......
	00 01 00 00 00 18 40 00 00 09 00 00 01 00 0f 57 | ......@........W
	6f 6c 66 54 50 4d 50 61 73 73 77 6f 72 64 00 20 | olfTPMPassword. 
	ee 93 83 f1 0b 6b ac 00 a4 66 12 bd 95 8f 9e 09 | .....k...f......
	b5 52 3c ec 7c 5f 8e 78 ec c9 49 e1 b1 57 82 fc | .R<.|_.x..I..W..
Response: 10
	80 01 00 00 00 0a 00 00 09 a2                   | ..........
TPM2_ObjectChangeAuth failed 0x9a2: TPM_RC_BAD_AUTH: Authorization failure without DA implications
TPM2_ObjectChangeAuth: private 2
Command: 14
	80 01 00 00 00 0e 00 00 01 65 80 00 00 02       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
Command: 12
	80 01 00 00 00 0c 00 00 01 78 00 03             | .........x..
Response: 227
	80 01 00 00 00 e3 00 00 00 00 00 03 01 00 00 20 | ............... 
	00 0b 00 10 00 20 ff ff ff ff 00 00 00 01 00 00 | ..... ..........
	00 00 00 00 00 00 00 00 00 00 ff ff ff ff ff ff | ................
	ff ff ff ff ff ff 00 20 ff ff ff ff 00 00 00 01 | ....... ........
	00 00 00 00 00 00 00 00 00 00 00 00 ff ff ff ff | ................
	ff ff ff ff ff ff ff fc 00 20 5a c6 35 d8 aa 3a | ......... Z.5..:
	93 e7 b3 eb bd 55 76 98 86 bc 65 1d 06 b0 cc 53 | .....Uv...e....S
	b0 f6 3b ce 3c 3e 27 d2 60 4b 00 20 6b 17 d1 f2 | ..;.<>'.`K. k...
	e1 2c 42 47 f8 bc e6 e5 63 a4 40 f2 77 03 7d 81 | .,BG....c.@.w.}.
	2d eb 33 a0 f4 a1 39 45 d8 98 c2 96 00 20 4f e3 | -.3...9E..... O.
	42 e2 fe 1a 7f 9b 8e e7 eb 4a 7c 0f 9e 16 2b ce | B........J|...+.
	33 57 6b 31 5e ce cb b6 40 68 37 bf 51 f5 00 20 | 3Wk1^...@h7.Q.. 
	ff ff ff ff 00 00 00 00 ff ff ff ff ff ff ff ff | ................
	bc e6 fa ad a7 17 9e 84 f3 b9 ca c2 fc 63 25 51 | .............c%Q
	00 01 01                                        | ...
TPM2_ECC_Parameters: CurveID 3, sz 256, p 32, a 32, b 32, gX 32, gY 32, n 32, h 1
Command: 102
	80 02 00 00 00 66 00 00 01 53 80 00 00 01 00 00 | .....f...S......
	00 18 40 00 00 09 00 00 01 00 0f 57 6f 6c 66 54 | ..@........WolfT
	50 4d 50 61 73 73 77 6f 72 64 00 1a 00 16 54 68 | PMPassword....Th
	69 73 49 73 41 53 65 63 72 65 74 55 73 61 67 65 | isIsASecretUsage
	41 75 74 68 00 00 00 18 00 23 00 0b 00 04 04 60 | Auth.....#.....`
	00 00 00 10 00 18 00 0b 00 03 00 10 00 00 00 00 | ................
	00 00 00 00 00 00                               | ......
Response: 428
	80 02 00 00 01 ac 00 00 00 00 00 00 01 99 00 7e | ...............~
	00 20 22 e2 bd 2e d0 a2 5e 4b b4 1d 11 10 95 3e | . ".....^K.....>
	0b b6 8e 6b c3 50 ef 4d 9d c4 a6 72 b6 72 2e c0 | ...k.P.M...r.r..
	91 77 00 10 04 07 2d da d1 0f ec 26 ae e1 5f 63 | .w....-....&.._c
	8c d6 15 c8 f9 b9 3d 8b d7 07 45 c5 38 dc ca e2 | ......=...E.8...
	c8 69 5b 52 f6 8e 97 43 3d 49 9f b4 12 5b 1b 2a | .i[R...C=I...[.*
	4d 31 64 06 de ec 6e 04 00 0f e8 2a 11 cf c1 ad | M1d...n....*....
	11 6e 0a 92 e2 d2 95 e4 2e 02 1b 42 45 83 19 93 | .n.........BE...
	50 13 44 2e b8 ef 54 f2 4d b1 f4 07 e7 e3 00 58 | P.D...T.M......X
	00 23 00 0b 00 04 04 60 00 00 00 10 00 18 00 0b | .#.....`........
	00 03 00 10 00 20 6a 90 50 e6 06 2d ec 40 92 6c | ..... j.P..-.@.l
	db b5 d2 36 09 0c bc b5 5c 98 56 49 06 71 77 ca | ...6....\.VI.qw.
	86 b2 ae 53 fe 6e 00 20 d4 0b a7 85 3e df 1d 91 | ...S.n. ....>...
	06 20 29 04 3c 60 9b a3 d2 b7 a5 53 a7 aa 52 4b | . ).<`.....S..RK
	f7 ea 2b 52 f1 69 88 df 00 73 00 00 00 00 00 20 | ..+R.i...s..... 
	e3 b0 c4 42 98 fc 1c 14 9a fb f4 c8 99 6f b9 24 | ...B.........o.$
	27 ae 41 e4 64 9b 93 4c a4 95 99 1b 78 52 b8 55 | '.A.d..L....xR.U
	01 00 0b 00 22 00 0b 23 24 f6 f5 3c 45 70 7c 3e | ...."..#$..<Ep|>
	0a d7 78 3e bc 01 ae cb d9 73 0a 54 49 92 14 63 | ..x>.....s.TI..c
	05 c6 a8 16 ed fc fc 00 22 00 0b 46 d7 59 62 84 | ........"..F.Yb.
	e1 eb e4 2d 48 d7 0c 35 33 ae 63 81 23 f2 e4 36 | ...-H..53.c.#..6
	e0 da 08 00 f4 6f 96 91 76 99 c8 00 00 00 20 31 | .....o..v..... 1
	13 9d 8e 56 51 c7 7b 71 e0 0c 82 ea 60 5b ca e3 | ...VQ.{q....`[..
	eb 54 32 ad f6 e1 9b 94 bb fa c9 b0 b6 8c 59 80 | .T2...........Y.
	21 40 00 00 01 00 20 28 f0 67 bd 33 54 3e 97 0e | !@.... (.g.3T>..
	38 f8 df db 99 14 92 b2 48 ac b6 98 74 2f 32 b2 | 8.......H...t/2.
	e0 e5 96 85 8a 04 ac 00 00 01 00 00             | ............
TPM2_Create: New ECDSA Key: pub 88, priv 126
Command: 260
	80 02 00 00 01 04 00 00 01 57 80 00 00 01 00 00 | .........W......
	00 18 40 00 00 09 00 00 01 00 0f 57 6f 6c 66 54 | ..@........WolfT
	50 4d 50 61 73 73 77 6f 72 64 00 7e 00 20 22 e2 | PMPassword.~. ".
	bd 2e d0 a2 5e 4b b4 1d 11 10 95 3e 0b b6 8e 6b | ....^K.....>...k
	c3 50 ef 4d 9d c4 a6 72 b6 72 2e c0 91 77 00 10 | .P.M...r.r...w..
	04 07 2d da d1 0f ec 26 ae e1 5f 63 8c d6 15 c8 | ..-....&.._c....
	f9 b9 3d 8b d7 07 45 c5 38 dc ca e2 c8 69 5b 52 | ..=...E.8....i[R
	f6 8e 97 43 3d 49 9f b4 12 5b 1b 2a 4d 31 64 06 | ...C=I...[.*M1d.
	de ec 6e 04 00 0f e8 2a 11 cf c1 ad 11 6e 0a 92 | ..n....*.....n..
	e2 d2 95 e4 2e 02 1b 42 45 83 19 93 50 13 44 2e | .......BE...P.D.
	b8 ef 54 f2 4d b1 f4 07 e7 e3 00 58 00 23 00 0b | ..T.M......X.#..
	00 04 04 60 00 00 00 10 00 18 00 0b 00 03 00 10 | ...`............
	00 20 6a 90 50 e6 06 2d ec 40 92 6c db b5 d2 36 | . j.P..-.@.l...6
	09 0c bc b5 5c 98 56 49 06 71 77 ca 86 b2 ae 53 | ....\.VI.qw....S
	fe 6e 00 20 d4 0b a7 85 3e df 1d 91 06 20 29 04 | .n. ....>.... ).
	3c 60 9b a3 d2 b7 a5 53 a7 aa 52 4b f7 ea 2b 52 | <`.....S..RK..+R
	f1 69 88 df                                     | .i..
Response: 59
	80 02 00 00 00 3b 00 00 00 00 80 00 00 02 00 00 | .....;..........
	00 24 00 22 00 0b e4 8c 73 c2 d2 43 a4 fd 61 73 | .$."....s..C..as
	66 ef 47 e7 19 ef d6 62 15 7a bd cb 8c e9 b1 30 | f.G....b.z.....0
	18 10 6f e8 7c 2c 00 00 01 00 00                | ..o.|,.....
TPM2_Load ECDSA Key Handle 0x80000002
Command: 95
	80 02 00 00 00 5f 00 00 01 5d 80 00 00 02 00 00 | ....._...]......
	00 1f 40 00 00 09 00 00 01 00 16 54 68 69 73 49 | ..@........ThisI
	73 41 53 65 63 72 65 74 55 73 61 67 65 41 75 74 | sASecretUsageAut
	68 00 20 11 11 11 11 11 11 11 11 11 11 11 11 11 | h. .............
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 00 18 00 0b 80 24 40 00 00 07 00 00    | ........$@.....
Response: 91
	80 02 00 00 00 5b 00 00 00 00 00 00 00 48 00 18 | .....[.......H..
	00 0b 00 20 30 16 51 f3 c1 a5 8e 19 8e ff a5 ae | ... 0.Q.........
	16 cc 8f 13 24 2b 5c 78 be 57 15 ac 1a cf e2 9d | ....$+\x.W......
	2b 90 ad 92 00 20 d3 cc f2 e4 43 c7 5b 78 f3 34 | +.... ....C.[x.4
	78 3e 9f 72 9c 54 db 1b e7 28 b3 56 b9 c3 b3 9e | x>.r.T...(.V....
	79 e2 3f 34 2a 45 00 00 01 00 00                | y.?4*E.....
TPM2_Sign: ECC S 32, R 32
Command: 120
	80 01 00 00 00 78 00 00 01 77 80 00 00 02 00 20 | .....x...w..... 
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	00 18 00 0b 00 20 30 16 51 f3 c1 a5 8e 19 8e ff | ..... 0.Q.......
	a5 ae 16 cc 8f 13 24 2b 5c 78 be 57 15 ac 1a cf | ......$+\x.W....
	e2 9d 2b 90 ad 92 00 20 d3 cc f2 e4 43 c7 5b 78 | ..+.... ....C.[x
	f3 34 78 3e 9f 72 9c 54 db 1b e7 28 b3 56 b9 c3 | .4x>.r.T...(.V..
	b3 9e 79 e2 3f 34 2a 45                         | ..y.?4*E
Response: 50
	80 01 00 00 00 32 00 00 00 00 80 22 40 00 00 01 | .....2....."@...
	00 20 df e2 fd 19 91 75 91 e3 8b 3f b0 5e d8 6d | . .....u...?.^.m
	f9 87 f9 a0 1a 95 3c b1 20 93 c9 ac be 3c 09 fb | ......<. ....<..
	9d 3a                                           | .:
TPM2_VerifySignature: Tag 32802
Command: 14
	80 01 00 00 00 0e 00 00 01 65 80 00 00 02       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
Command: 102
	80 02 00 00 00 66 00 00 01 53 80 00 00 01 00 00 | .....f...S......
	00 18 40 00 00 09 00 00 01 00 0f 57 6f 6c 66 54 | ..@........WolfT
	50 4d 50 61 73 73 77 6f 72 64 00 1a 00 16 54 68 | PMPassword....Th
	69 73 49 73 41 53 65 63 72 65 74 55 73 61 67 65 | isIsASecretUsage
	41 75 74 68 00 00 00 18 00 23 00 0b 00 02 04 60 | Auth.....#.....`
	00 00 00 10 00 19 00 0b 00 03 00 10 00 00 00 00 | ................
	00 00 00 00 00 00                               | ......
Response: 428
	80 02 00 00 01 ac 00 00 00 00 00 00 01 99 00 7e | ...............~
	00 20 b6 46 c8 82 6b 8c 7a 4b e4 1b 96 fa 21 1b | . .F..k.zK....!.
	0d 08 9e c3 32 22 9a 3f 41 3c 2d ae 07 59 73 74 | ....2".?A<-..Yst
	42 3f 00 10 10 41 4c 5e 55 84 67 ce b5 b8 04 91 | B?...AL^U.g.....
	f7 c6 55 ab ee 13 0b f4 07 cb 5c 66 1c 74 6c f2 | ..U.......\f.tl.
	e9 87 93 94 78 79 50 49 85 f8 b3 6a 5c 4a 2d 50 | ....xyPI...j\J-P
	d8 6c a8 26 5c 89 b6 3e 2c 61 20 6b fb 88 ae 1f | .l.&\..>,a k....
	60 47 39 2f d2 ab 12 bd 1d ef 91 e4 7e c5 89 f8 | `G9/........~...
	b8 d0 f5 15 f5 e8 2f c2 ea 34 de fd ac bb 00 58 | ....../..4.....X
	00 23 00 0b 00 02 04 60 00 00 00 10 00 19 00 0b | .#.....`........
	00 03 00 10 00 20 0a ef ef 11 1c 75 ef 4b ee 10 | ..... .....u.K..
	de bc 51 13 6e ef 82 28 3c ac 71 bf cc b7 7c a1 | ..Q.n..(<.q...|.
	e5 d5 c3 54 5d fb 00 20 d0 51 6e f3 06 47 9b dc | ...T].. .Qn..G..
	67 94 a7 ac 6f 97 73 80 93 af 75 0f 50 17 7f 1f | g...o.s...u.P...
	9b ec 88 77 4e 7c eb b3 00 73 00 00 00 00 00 20 | ...wN|...s..... 
	e3 b0 c4 42 98 fc 1c 14 9a fb f4 c8 99 6f b9 24 | ...B.........o.$
	27 ae 41 e4 64 9b 93 4c a4 95 99 1b 78 52 b8 55 | '.A.d..L....xR.U
	01 00 0b 00 22 00 0b 23 24 f6 f5 3c 45 70 7c 3e | ...."..#$..<Ep|>
	0a d7 78 3e bc 01 ae cb d9 73 0a 54 49 92 14 63 | ..x>.....s.TI..c
	05 c6 a8 16 ed fc fc 00 22 00 0b 46 d7 59 62 84 | ........"..F.Yb.
	e1 eb e4 2d 48 d7 0c 35 33 ae 63 81 23 f2 e4 36 | ...-H..53.c.#..6
	e0 da 08 00 f4 6f 96 91 76 99 c8 00 00 00 20 31 | .....o..v..... 1
	13 9d 8e 56 51 c7 7b 71 e0 0c 82 ea 60 5b ca e3 | ...VQ.{q....`[..
	eb 54 32 ad f6 e1 9b 94 bb fa c9 b0 b6 8c 59 80 | .T2...........Y.
	21 40 00 00 01 00 20 15 27 80 ce ef a6 4f 1b 7c | !@.... .'....O.|
	87 d2 79 39 cc 63 89 31 18 0b 67 f2 5d 67 4a a5 | ..y9.c.1..g.]gJ.
	1e ec 13 20 75 15 a3 00 00 01 00 00             | ... u.......
TPM2_Create: New ECDH Key: pub 88, priv 126
Command: 260
	80 02 00 00 01 04 00 00 01 57 80 00 00 01 00 00 | .........W......
	00 18 40 00 00 09 00 00 01 00 0f 57 6f 6c 66 54 | ..@........WolfT
	50 4d 50 61 73 73 77 6f 72 64 00 7e 00 20 b6 46 | PMPassword.~. .F
	c8 82 6b 8c 7a 4b e4 1b 96 fa 21 1b 0d 08 9e c3 | ..k.zK....!.....
	32 22 9a 3f 41 3c 2d ae 07 59 73 74 42 3f 00 10 | 2".?A<-..YstB?..
	10 41 4c 5e 55 84 67 ce b5 b8 04 91 f7 c6 55 ab | .AL^U.g.......U.
	ee 13 0b f4 07 cb 5c 66 1c 74 6c f2 e9 87 93 94 | ......\f.tl.....
	78 79 50 49 85 f8 b3 6a 5c 4a 2d 50 d8 6c a8 26 | xyPI...j\J-P.l.&
	5c 89 b6 3e 2c 61 20 6b fb 88 ae 1f 60 47 39 2f | \..>,a k....`G9/
	d2 ab 12 bd 1d ef 91 e4 7e c5 89 f8 b8 d0 f5 15 | ........~.......
	f5 e8 2f c2 ea 34 de fd ac bb 00 58 00 23 00 0b | ../..4.....X.#..
	00 02 04 60 00 00 00 10 00 19 00 0b 00 03 00 10 | ...`............
	00 20 0a ef ef 11 1c 75 ef 4b ee 10 de bc 51 13 | . .....u.K....Q.
	6e ef 82 28 3c ac 71 bf cc b7 7c a1 e5 d5 c3 54 | n..(<.q...|....T
	5d fb 00 20 d0 51 6e f3 06 47 9b dc 67 94 a7 ac | ].. .Qn..G..g...
	6f 97 73 80 93 af 75 0f 50 17 7f 1f 9b ec 88 77 | o.s...u.P......w
	4e 7c eb b3                                     | N|..
Response: 59
	80 02 00 00 00 3b 00 00 00 00 80 00 00 02 00 00 | .....;..........
	00 24 00 22 00 0b ae 32 6f 38 af a0 61 ff 6d 6f | .$."...2o8..a.mo
	cf 22 10 a1 71 b3 bc 7f 40 d2 50 d8 b1 48 43 45 | ."..q...@.P..HCE
	6e b3 f3 cd e7 da 00 00 01 00 00                | n..........
TPM2_Load ECDH Key Handle 0x80000002
Command: 14
	80 01 00 00 00 0e 00 00 01 63 80 00 00 02       | .........c....
Response: 150
	80 01 00 00 00 96 00 00 00 00 00 44 00 20 71 3c | ...........D. q<
	7b 00 6a ba 55 99 7d 50 0c 01 10 e2 a4 52 ee 7b | {.j.U.}P.....R.{
	7a a9 d0 f4 4d bc 9c bb d0 90 c1 c9 b7 ff 00 20 | z...M.......... 
	50 5f 67 e2 c6 6b 3b 79 70 e6 89 bd 10 00 84 71 | P_g..k;yp......q
	26 de b9 94 65 5e 25 2f be 53 03 0c b9 93 19 f0 | &...e^%/.S......
	00 44 00 20 f5 7e 91 af 96 2f 85 60 2b 57 f2 1d | .D. .~.../.`+W..
	d1 b6 13 4c 46 7e 58 b7 33 44 6a 89 7b cc 2e 76 | ...LF~X.3Dj.{..v
	cb 77 11 6a 00 20 91 f0 72 fc d6 c8 93 68 0a 93 | .w.j. ..r....h..
	22 83 29 d9 71 a6 7f eb 6c 73 a8 74 08 dc 0e 38 | ".).q...ls.t...8
	ec 2f 67 11 87 d3                               | ./g...
TPM2_ECDH_KeyGen: zPt 68, pubPt 68
Command: 14
	80 01 00 00 00 0e 00 00 01 65 80 00 00 02       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
Command: 387
	80 02 00 00 01 83 00 00 01 53 80 00 00 01 00 00 | .........S......
	00 18 40 00 00 09 00 00 01 00 0f 57 6f 6c 66 54 | ..@........WolfT
	50 4d 50 61 73 73 77 6f 72 64 00 1a 00 16 54 68 | PMPassword....Th
	69 73 49 73 41 53 65 63 72 65 74 55 73 61 67 65 | isIsASecretUsage
	41 75 74 68 00 00 01 16 00 01 00 0b 00 06 04 60 | Auth...........`
	00 00 00 10 00 10 08 00 00 00 00 00 01 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 | ................
	00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 1f | ................
	52 61 6e 64 6f 6d 53 65 72 76 65 72 50 69 63 6b | RandomServerPick
	65 64 43 72 65 61 74 69 6f 6e 4e 6f 6e 63 65 00 | edCreationNonce.
	00 00 00                                        | ...
Response: 745
	80 02 00 00 02 e9 00 00 00 00 00 00 02 d6 00 de | ................
	00 20 6a 07 24 9f 5c 91 6e ce 31 5b e5 08 7d e0 | . j.$.\.n.1[..}.
	3f 0a ec 7b 91 db 5e c1 74 c9 b6 49 d3 09 27 cc | ?..{..^.t..I..'.
	66 ca 00 10 27 9b 7e 1d 7a e3 da 97 d9 70 20 84 | f...'.~.z....p .
	16 bc 45 07 80 fc f5 db 23 99 26 a1 19 54 fe bb | ..E.....#.&..T..
	41 d8 43 bf 2f c6 93 08 bb 57 b0 02 31 f3 2f 2e | A.C./....W..1./.
	81 0a 41 d5 f6 8d e4 0b 16 34 b1 1d 53 e4 01 11 | ..A......4..S...
	e9 c7 19 83 3a 68 c7 0a 41 b9 72 78 e0 99 e7 26 | ....:h..A.rx...&
	21 89 04 70 47 f6 2d c3 8b b7 29 7c ba 2c 79 8c | !..pG.-...)|.,y.
	90 99 5d f2 42 d6 6a 0b b7 da 2d e4 e9 85 c9 c4 | ..].B.j...-.....
	b4 d8 49 f0 7d 29 e4 ff d5 c8 60 b6 43 d8 72 2f | ..I.})....`.C.r/
	6e 15 27 a3 83 d5 f4 32 e8 59 5b d0 17 f2 cf 51 | n.'....2.Y[....Q
	69 4c 2e e2 90 cd 3f bb f0 ce 18 91 f3 e1 6f cf | iL....?.......o.
	09 f4 42 b1 04 23 f5 02 ca 8e 96 07 25 9a f1 aa | ..B..#......%...
	c1 11 b9 c9 39 33 cd ab c6 42 d5 af 48 ad 01 16 | ....93...B..H...
	00 01 00 0b 00 06 04 60 00 00 00 10 00 10 08 00 | .......`........
	00 00 00 00 01 00 ee 54 bd bf 29 8d ca 81 02 c8 | .......T..).....
	22 69 17 11 ef a0 19 ea f7 a1 60 30 3b 30 61 a0 | "i........`0;0a.
	2b ba ed 9a 84 f0 07 52 38 7c b5 56 23 62 1b 33 | +......R8|.V#b.3
	e6 97 d5 f1 45 e8 06 14 3c 39 70 c9 0c d4 29 f8 | ....E...<9p...).
	89 93 d1 42 d3 b3 0d 14 79 e9 ee c4 fd 53 d8 fd | ...B....y....S..
	0f d5 e9 e6 ac 71 aa ec f2 d4 9a 23 5a 68 3e a4 | .....q.....#Zh>.
	b9 ce 8b bf a6 48 41 97 d6 9a 23 db e8 42 a1 8a | .....HA...#..B..
	08 26 82 80 63 f2 60 8b c8 b5 0a 9c eb bb b4 56 | .&..c.`........V
	aa 4a b3 01 f1 41 4a fa c0 0a 49 46 59 fa 95 a9 | .J...AJ...IFY...
	64 3b c2 fa b0 8e bf aa b6 74 ee f9 c9 3a d8 bb | d;.......t...:..
	e9 7e 7e a6 66 cb 76 b9 0a ba b5 fc f8 4f 5e de | .~~.f.v......O^.
	da 5b 88 da 27 55 4b 4d 2f bd 94 52 05 a1 d7 f8 | .[..'UKM/..R....
	89 26 41 5e 3e 24 4a 1f b0 7c 69 7f 0f e8 b7 98 | .&A^>$J..|i.....
	08 21 5e ed 82 74 3c b7 2a 09 d5 aa c6 d6 e5 a3 | .!^..t<.*.......
	7f b2 83 fe ce 21 d6 ab 36 77 a8 a0 d1 7f 7b 22 | .....!..6w....{"
	37 19 5d a3 3e 45 c3 a7 7c 59 e0 4a 5a 4d 84 aa | 7.].>E..|Y.JZM..
	b4 bd 22 a0 34 97 00 92 00 00 00 00 00 20 e3 b0 | ..".4........ ..
	c4 42 98 fc 1c 14 9a fb f4 c8 99 6f b9 24 27 ae | .B.........o.$'.
	41 e4 64 9b 93 4c a4 95 99 1b 78 52 b8 55 01 00 | A.d..L....xR.U..
	0b 00 22 00 0b 23 24 f6 f5 3c 45 70 7c 3e 0a d7 | .."..#$..<Ep|>..
	78 3e bc 01 ae cb d9 73 0a 54 49 92 14 63 05 c6 | x>.....s.TI..c..
	a8 16 ed fc fc 00 22 00 0b 46 d7 59 62 84 e1 eb | ......"..F.Yb...
	e4 2d 48 d7 0c 35 33 ae 63 81 23 f2 e4 36 e0 da | .-H..53.c.#..6..
	08 00 f4 6f 96 91 76 99 c8 00 1f 52 61 6e 64 6f | ...o..v....Rando
	6d 53 65 72 76 65 72 50 69 63 6b 65 64 43 72 65 | mServerPickedCre
	61 74 69 6f 6e 4e 6f 6e 63 65 00 20 1a 5f 91 c1 | ationNonce. ._..
	77 a6 20 8a 1d 8a fa 98 7a e1 4e b2 01 6a 99 22 | w. .....z.N..j."
	c8 a9 35 e4 d8 72 a5 c2 7b 2a cf 21 80 21 40 00 | ..5..r..{*.!.!@.
	00 01 00 20 55 72 78 4e 7b 7e a7 3a a3 6d 59 98 | ... UrxN{~.:.mY.
	54 a8 06 af 5c a3 40 ea 0f b9 be 9f 91 06 3a 68 | T...\.@.......:h
	85 99 71 35 00 00 01 00 00                      | ..q5.....
TPM2_Create: New RSA Key: pub 278, priv 222
Command: 546
	80 02 00 00 02 22 00 00 01 57 80 00 00 01 00 00 | ....."...W......
	00 18 40 00 00 09 00 00 01 00 0f 57 6f 6c 66 54 | ..@........WolfT
	50 4d 50 61 73 73 77 6f 72 64 00 de 00 20 6a 07 | PMPassword... j.
	24 9f 5c 91 6e ce 31 5b e5 08 7d e0 3f 0a ec 7b | $.\.n.1[..}.?..{
	91 db 5e c1 74 c9 b6 49 d3 09 27 cc 66 ca 00 10 | ..^.t..I..'.f...
	27 9b 7e 1d 7a e3 da 97 d9 70 20 84 16 bc 45 07 | '.~.z....p ...E.
	80 fc f5 db 23 99 26 a1 19 54 fe bb 41 d8 43 bf | ....#.&..T..A.C.
	2f c6 93 08 bb 57 b0 02 31 f3 2f 2e 81 0a 41 d5 | /....W..1./...A.
	f6 8d e4 0b 16 34 b1 1d 53 e4 01 11 e9 c7 19 83 | .....4..S.......
	3a 68 c7 0a 41 b9 72 78 e0 99 e7 26 21 89 04 70 | :h..A.rx...&!..p
	47 f6 2d c3 8b b7 29 7c ba 2c 79 8c 90 99 5d f2 | G.-...)|.,y...].
	42 d6 6a 0b b7 da 2d e4 e9 85 c9 c4 b4 d8 49 f0 | B.j...-.......I.
	7d 29 e4 ff d5 c8 60 b6 43 d8 72 2f 6e 15 27 a3 | })....`.C.r/n.'.
	83 d5 f4 32 e8 59 5b d0 17 f2 cf 51 69 4c 2e e2 | ...2.Y[....QiL..
	90 cd 3f bb f0 ce 18 91 f3 e1 6f cf 09 f4 42 b1 | ..?.......o...B.
	04 23 f5 02 ca 8e 96 07 25 9a f1 aa c1 11 b9 c9 | .#......%.......
	39 33 cd ab c6 42 d5 af 48 ad 01 16 00 01 00 0b | 93...B..H.......
	00 06 04 60 00 00 00 10 00 10 08 00 00 00 00 00 | ...`............
	01 00 ee 54 bd bf 29 8d ca 81 02 c8 22 69 17 11 | ...T..)....."i..
	ef a0 19 ea f7 a1 60 30 3b 30 61 a0 2b ba ed 9a | ......`0;0a.+...
	84 f0 07 52 38 7c b5 56 23 62 1b 33 e6 97 d5 f1 | ...R8|.V#b.3....
	45 e8 06 14 3c 39 70 c9 0c d4 29 f8 89 93 d1 42 | E...<9p...)....B
	d3 b3 0d 14 79 e9 ee c4 fd 53 d8 fd 0f d5 e9 e6 | ....y....S......
	ac 71 aa ec f2 d4 9a 23 5a 68 3e a4 b9 ce 8b bf | .q.....#Zh>.....
	a6 48 41 97 d6 9a 23 db e8 42 a1 8a 08 26 82 80 | .HA...#..B...&..
	63 f2 60 8b c8 b5 0a 9c eb bb b4 56 aa 4a b3 01 | c.`........V.J..
	f1 41 4a fa c0 0a 49 46 59 fa 95 a9 64 3b c2 fa | .AJ...IFY...d;..
	b0 8e bf aa b6 74 ee f9 c9 3a d8 bb e9 7e 7e a6 | .....t...:...~~.
	66 cb 76 b9 0a ba b5 fc f8 4f 5e de da 5b 88 da | f.v......O^..[..
	27 55 4b 4d 2f bd 94 52 05 a1 d7 f8 89 26 41 5e | 'UKM/..R.....&A^
	3e 24 4a 1f b0 7c 69 7f 0f e8 b7 98 08 21 5e ed | >$J..|i......!^.
	82 74 3c b7 2a 09 d5 aa c6 d6 e5 a3 7f b2 83 fe | .t<.*...........
	ce 21 d6 ab 36 77 a8 a0 d1 7f 7b 22 37 19 5d a3 | .!..6w....{"7.].
	3e 45 c3 a7 7c 59 e0 4a 5a 4d 84 aa b4 bd 22 a0 | >E..|Y.JZM....".
	34 97                                           | 4.
Response: 59
	80 02 00 00 00 3b 00 00 00 00 80 00 00 02 00 00 | .....;..........
	00 24 00 22 00 0b 42 90 b5 f7 42 54 b6 c9 22 af | .$."..B...BT..".
	0b e1 f4 3d b8 68 d7 90 b3 62 a8 f9 4d 3f 51 90 | ...=.h...b..M?Q.
	d5 07 6a 8d 2a fc 00 00 01 00 00                | ..j.*......
TPM2_Load RSA Key Handle 0x80000002
Command: 68
	80 01 00 00 00 44 00 00 01 74 80 00 00 02 00 20 | .....D...t..... 
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	00 17 00 0b 00 0e 54 68 69 73 49 73 4d 79 4c 61 | ......ThisIsMyLa
	62 65 6c 00                                     | bel.
Response: 268
	80 01 00 00 01 0c 00 00 00 00 01 00 e0 fa bc c6 | ................
	0e 9c 3c b5 05 6f b7 e2 f1 e2 f7 6d ce 23 c4 22 | ..<..o.....m.#."
	d4 2c 43 88 cc 36 99 0f c2 27 45 21 7c 4b 0c 51 | .,C..6...'E!|K.Q
	af 54 52 66 bc 46 4c 9f e6 0f 6a b6 66 18 8c 19 | .TRf.FL...j.f...
	3f 98 c7 69 2f 82 3d b1 a4 db ff 17 28 e7 ab 94 | ?..i/.=.....(...
	bd 9b 86 5e cd 5f 4a e5 41 a4 14 c9 d2 d2 f7 ee | ...^._J.A.......
	d6 54 54 b8 45 ec c9 0b ae 1a 64 e9 4b 82 72 f3 | .TT.E.....d.K.r.
	48 28 d8 b2 f1 13 aa 8d 7a bf fd 5c 6e b5 ae e4 | H(......z..\n...
	80 74 0d a7 3c 9d 11 1f 1f ee ee 8f 63 36 3f 7a | .t..<.......c6?z
	23 ca b1 7a 5e 00 d0 94 c7 2e 85 b9 6f 4c 40 8a | #..z^.......oL@.
	d0 11 91 9b 70 71 c1 b5 dc 3e 3f b8 91 60 48 7a | ....pq...>?..`Hz
	ed 92 99 49 1b 68 77 26 e6 32 1c f7 a0 6f 30 2f | ...I.hw&.2...o0/
	b5 d0 fd 48 f9 4c d6 fb f7 da d4 3a 64 b8 0d a4 | ...H.L.....:d...
	31 61 38 62 f4 30 86 de ca ee 37 49 21 76 88 9d | 1a8b.0....7I!v..
	19 0d 8a ca 1c 5f 79 c9 36 5c cc 61 ae 73 81 d6 | ....._y.6\.a.s..
	74 59 38 3d 51 be 28 3a ff 20 a3 17 0e a6 2f b2 | tY8=Q.(:. ..../.
	34 76 90 1d 5b f6 04 f4 cf 80 14 73             | 4v..[......s
TPM2_RSA_Encrypt: 256
Command: 327
	80 02 00 00 01 47 00 00 01 59 80 00 00 02 00 00 | .....G...Y......
	00 1f 40 00 00 09 00 00 01 00 16 54 68 69 73 49 | ..@........ThisI
	73 41 53 65 63 72 65 74 55 73 61 67 65 41 75 74 | sASecretUsageAut
	68 01 00 e0 fa bc c6 0e 9c 3c b5 05 6f b7 e2 f1 | h........<..o...
	e2 f7 6d ce 23 c4 22 d4 2c 43 88 cc 36 99 0f c2 | ..m.#.".,C..6...
	27 45 21 7c 4b 0c 51 af 54 52 66 bc 46 4c 9f e6 | 'E!|K.Q.TRf.FL..
	0f 6a b6 66 18 8c 19 3f 98 c7 69 2f 82 3d b1 a4 | .j.f...?..i/.=..
	db ff 17 28 e7 ab 94 bd 9b 86 5e cd 5f 4a e5 41 | ...(......^._J.A
	a4 14 c9 d2 d2 f7 ee d6 54 54 b8 45 ec c9 0b ae | ........TT.E....
	1a 64 e9 4b 82 72 f3 48 28 d8 b2 f1 13 aa 8d 7a | .d.K.r.H(......z
	bf fd 5c 6e b5 ae e4 80 74 0d a7 3c 9d 11 1f 1f | ..\n....t..<....
	ee ee 8f 63 36 3f 7a 23 ca b1 7a 5e 00 d0 94 c7 | ...c6?z#..z^....
	2e 85 b9 6f 4c 40 8a d0 11 91 9b 70 71 c1 b5 dc | ...oL@.....pq...
	3e 3f b8 91 60 48 7a ed 92 99 49 1b 68 77 26 e6 | >?..`Hz...I.hw&.
	32 1c f7 a0 6f 30 2f b5 d0 fd 48 f9 4c d6 fb f7 | 2...o0/...H.L...
	da d4 3a 64 b8 0d a4 31 61 38 62 f4 30 86 de ca | ..:d...1a8b.0...
	ee 37 49 21 76 88 9d 19 0d 8a ca 1c 5f 79 c9 36 | .7I!v......._y.6
	5c cc 61 ae 73 81 d6 74 59 38 3d 51 be 28 3a ff | \.a.s..tY8=Q.(:.
	20 a3 17 0e a6 2f b2 34 76 90 1d 5b f6 04 f4 cf |  ..../.4v..[....
	80 14 73 00 17 00 0b 00 0e 54 68 69 73 49 73 4d | ..s......ThisIsM
	79 4c 61 62 65 6c 00                            | yLabel.
Response: 53
	80 02 00 00 00 35 00 00 00 00 00 00 00 22 00 20 | .....5.......". 
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 | ................
	00 00 01 00 00                                  | .....
TPM2_RSA_Decrypt: 32
RSA Encrypt/Decrypt test passed
Command: 14
	80 01 00 00 00 0e 00 00 01 65 80 00 00 02       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
Command: 14
	80 01 00 00 00 0e 00 00 01 69 01 00 00 00       | .........i....
Response: 10
	80 01 00 00 00 0a 00 00 01 8b                   | ..........
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE: The handle is not correct for the use
Command: 14
	80 01 00 00 00 0e 00 00 01 69 01 00 00 01       | .........i....
Response: 10
	80 01 00 00 00 0a 00 00 01 8b                   | ..........
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE: The handle is not correct for the use
Command: 14
	80 01 00 00 00 0e 00 00 01 69 01 00 00 02       | .........i....
Response: 10
	80 01 00 00 00 0a 00 00 01 8b                   | ..........
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE: The handle is not correct for the use
Command: 14
	80 01 00 00 00 0e 00 00 01 69 01 00 00 03       | .........i....
Response: 10
	80 01 00 00 00 0a 00 00 01 8b                   | ..........
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE: The handle is not correct for the use
Command: 14
	80 01 00 00 00 0e 00 00 01 69 01 00 00 04       | .........i....
Response: 10
	80 01 00 00 00 0a 00 00 01 8b                   | ..........
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE: The handle is not correct for the use
Command: 14
	80 01 00 00 00 0e 00 00 01 69 01 00 00 05       | .........i....
Response: 10
	80 01 00 00 00 0a 00 00 01 8b                   | ..........
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE: The handle is not correct for the use
Command: 14
	80 01 00 00 00 0e 00 00 01 69 01 00 00 06       | .........i....
Response: 10
	80 01 00 00 00 0a 00 00 01 8b                   | ..........
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE: The handle is not correct for the use
Command: 14
	80 01 00 00 00 0e 00 00 01 69 01 00 00 07       | .........i....
Response: 10
	80 01 00 00 00 0a 00 00 01 8b                   | ..........
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE: The handle is not correct for the use
Command: 14
	80 01 00 00 00 0e 00 00 01 69 01 00 00 08       | .........i....
Response: 10
	80 01 00 00 00 0a 00 00 01 8b                   | ..........
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE: The handle is not correct for the use
Command: 14
	80 01 00 00 00 0e 00 00 01 69 01 00 00 09       | .........i....
Response: 10
	80 01 00 00 00 0a 00 00 01 8b                   | ..........
TPM2_NV_ReadPublic failed 0x18b: TPM_RC_HANDLE: The handle is not correct for the use
Command: 82
	80 02 00 00 00 52 00 00 01 2a 80 00 00 01 00 00 | .....R...*......
	00 18 40 00 00 09 00 00 01 00 0f 57 6f 6c 66 54 | ..@........WolfT
	50 4d 50 61 73 73 77 6f 72 64 00 16 54 68 69 73 | PMPassword..This
	49 73 41 53 65 63 72 65 74 55 73 61 67 65 41 75 | IsASecretUsageAu
	74 68 00 0e 01 bf ff ff 00 0b 06 02 00 02 00 00 | th..............
	00 20                                           | . 
Response: 10
	80 01 00 00 00 0a 00 00 01 84                   | ..........
TPM2_NV_DefineSpace failed 0x184: TPM_RC_VALUE: Value is out of range or is not correct for the context
Command: 14
	80 01 00 00 00 0e 00 00 01 65 03 00 00 00       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
Command: 14
	80 01 00 00 00 0e 00 00 01 65 80 00 00 02       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 01 cb                   | ..........
Command: 14
	80 01 00 00 00 0e 00 00 01 65 80 00 00 02       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 01 cb                   | ..........
Command: 14
	80 01 00 00 00 0e 00 00 01 65 80 00 00 02       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 01 cb                   | ..........
Command: 14
	80 01 00 00 00 0e 00 00 01 65 80 00 00 02       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 01 cb                   | ..........
Command: 14
	80 01 00 00 00 0e 00 00 01 65 80 00 00 00       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
Command: 14
	80 01 00 00 00 0e 00 00 01 65 80 00 00 01       | .........e....
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
Command: 12
	80 01 00 00 00 0c 00 00 01 45 00 00             | .........E..
Response: 10
	80 01 00 00 00 0a 00 00 00 00                   | ..........
```


## Todo
* Improve overall documentation.
* Add support for using the TPM with wolfSSL for TLS.
* Add support for encrypting / decrypting parameters. (90% complete)
* Add `spi_tis_dev` support for Raspberry Pi. This will allow use with Rasbian kernel patches with Infineon.


## Support

Email us at [support@wolfssl.com](mailto:support@wolfssl.com).